from __future__ import absolute_import, division, print_function

import sys
from types import MethodType

import pandas.util.testing as tm
from toolz import identity

PY3 = sys.version_info[0] == 3
PY2 = sys.version_info[0] == 2

if PY3:
    import builtins

    def apply(f, args, **kwargs):
        return f(*args, **kwargs)
else:
    import __builtin__ as builtins
    apply = builtins.apply


try:
    import cPickle as pickle
except ImportError:
    import pickle


# Portions of this taken from the six library, licensed as follows.
#
# Copyright (c) 2010-2013 Benjamin Peterson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from toolz.compatibility import map, zip, range, reduce


if PY2:
    _inttypes = (int, long)
    unicode = builtins.unicode
    basestring = builtins.basestring
    _strtypes = (str, unicode)

    def boundmethod(func, instance):
        return MethodType(func, instance, type(instance))

    from itertools import izip_longest as zip_longest
    from contextlib2 import ExitStack
else:
    _inttypes = (int,)
    _strtypes = (str,)
    unicode = str
    basestring = str
    boundmethod = MethodType

    from itertools import zip_longest
    from contextlib import ExitStack


import io


try:
    SEEK_END = io.SEEK_END
except AttributeError:
    SEEK_END = 2


try:
    import pytest
    skipif = pytest.mark.skipif
    xfail = pytest.mark.xfail
    min_python_version = skipif(sys.version_info < (2, 7),
                                reason="Python >= 2.7 required")
    raises = pytest.raises
except ImportError:
    # TODO: move the above into a separate testing utils module
    pass


if sys.version_info >= (2, 7):
    from ctypes import c_ssize_t
else:
    import ctypes
    if ctypes.sizeof(ctypes.c_void_p) == 4:
        c_ssize_t = ctypes.c_int32
    else:
        c_ssize_t = ctypes.c_int64


try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


def assert_series_equal(left, right, check_names=True, **kwargs):
    """Backwards compatibility wrapper for
    ``pandas.util.testing.assert_series_equal``

    Examples
    --------
    >>> import pandas as pd
    >>> s = pd.Series(list('abc'), name='a')
    >>> s2 = pd.Series(list('abc'), name='b')
    >>> assert_series_equal(s, s2)  # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    AssertionError: ...
    >>> assert_series_equal(s, s2, check_names=False)

    See Also
    --------
    pandas.util.testing.assert_series_equal
    """
    try:
        return tm.assert_series_equal(left, right, check_names=check_names,
                                      **kwargs)
    except TypeError:
        if check_names:
            assert left.name == right.name
        return tm.assert_series_equal(left, right, **kwargs)


if PY2:
    def u8(cs):
        return cs.decode('utf-8')
else:
    u8 = identity



from __future__ import absolute_import, division, print_function

from ..expr import (Expr, Symbol, Field, Arithmetic, UnaryMath, Not, USub,
                    isnan, UnaryOp, BinOp)
from toolz import curry
import itertools
from ..expr.broadcast import broadcast_collect


funcnames = ('func_%d' % i for i in itertools.count())


def parenthesize(s):
    if ' ' in s:
        return '(%s)' % s
    else:
        return s


def print_numexpr(leaves, expr):
    """ Print expression to be evaluated in Python

    >>> from blaze import symbol, ceil, sin, cos

    >>> t = symbol('t', 'var * {x: int, y: int, z: int, when: datetime}')
    >>> print_numexpr([t], t.x + t.y)
    'x + y'

    Supports mathematical functions

    >>> print_numexpr([t], sin(t.x) > cos(t.y))
    'sin(x) > cos(y)'

    Returns
    -------

    s: string
       A evalable string
    """
    if not isinstance(expr, Expr):
        return repr(expr)
    if any(expr.isidentical(leaf) for leaf in leaves):
        return expr._name
    if isinstance(expr, Symbol):
        return expr._name
    if isinstance(expr, Field):
        return expr._name
    if isinstance(expr, Arithmetic):
        lhs = print_numexpr(leaves, expr.lhs)
        rhs = print_numexpr(leaves, expr.rhs)
        return '%s %s %s' % (parenthesize(lhs),
                             expr.symbol,
                             parenthesize(rhs))
    if isinstance(expr, UnaryMath):
        child = print_numexpr(leaves, expr._child)
        return '%s(%s)' % (type(expr).__name__, child)
    if isinstance(expr, UnaryOp) and hasattr(expr, 'symbol'):
        child = print_numexpr(leaves, expr._child)
        return '%s%s' % (expr.symbol, parenthesize(child))
    if isinstance(expr, isnan):
        child = print_numexpr(leaves, expr._child)
        return '%s != %s' % (parenthesize(child), parenthesize(child))
    raise NotImplementedError("Operation %s not supported by numexpr" %
                              type(expr).__name__)


Broadcastable = WantToBroadcast = BinOp, UnaryOp

broadcast_numexpr_collect = curry(
    broadcast_collect,
    broadcastable=Broadcastable,
    want_to_broadcast=WantToBroadcast
)



from __future__ import absolute_import, print_function, division

from getpass import getuser

import pytest

sa = pytest.importorskip('sqlalchemy')
pytest.importorskip('pymysql')

from odo import odo, drop, discover

import pandas as pd
import numpy as np

from blaze import symbol, compute
from blaze.utils import example, normalize
from blaze.interactive import iscoretype, iscorescalar, iscoresequence


@pytest.yield_fixture(scope='module')
def data():
    try:
        t = odo(
            example('nyc.csv'),
            'mysql+pymysql://%s@localhost/test::nyc' % getuser()
        )
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield t.bind
        finally:
            drop(t)


@pytest.fixture
def db(data):
    return symbol('test', discover(data))


def test_agg_sql(db, data):
    subset = db.nyc[['pickup_datetime', 'dropoff_datetime', 'passenger_count']]
    expr = subset[subset.passenger_count < 4].passenger_count.min()
    result = compute(expr, data, return_type='native')
    expected = """
    select
        min(alias.passenger_count) as passenger_count_min
    from
        (select
            nyc.passenger_count as passenger_count
         from
            nyc
         where nyc.passenger_count < % ( passenger_count_1 ) s) as alias
    """
    assert normalize(str(result)) == normalize(expected)


def test_agg_compute(db, data):
    subset = db.nyc[['pickup_datetime', 'dropoff_datetime', 'passenger_count']]
    expr = subset[subset.passenger_count < 4].passenger_count.min()
    result = compute(expr, data, return_type='native')
    passenger_count = odo(compute(db.nyc.passenger_count, {db: data}, return_type='native'), pd.Series)
    assert passenger_count[passenger_count < 4].min() == result.scalar()


def test_core_compute(db, data):
    assert isinstance(compute(db.nyc, data, return_type='core'), pd.DataFrame)
    assert isinstance(compute(db.nyc.passenger_count, data, return_type='core'), pd.Series)
    assert iscorescalar(compute(db.nyc.passenger_count.mean(), data, return_type='core'))
    assert isinstance(compute(db.nyc, data, return_type=list), list)



from blaze import symbol
from blaze.expr.math import abs as mathabs, sin

x = symbol('x', '5 * 3 * int')


def test_math_shapes():
    assert sin(x).shape == x.shape


def test_abs():
    assert abs(x).isidentical(mathabs(x))



from __future__ import absolute_import, division, print_function

import pytest
bcolz = pytest.importorskip('bcolz')

from datashape import discover, dshape

import numpy as np

import pandas.util.testing as tm

from odo import into
from blaze import by
from blaze.expr import symbol
from blaze.compute.core import compute, pre_compute
from blaze.compute.bcolz import get_chunksize


b = bcolz.ctable(np.array([(1, 1., np.datetime64('2010-01-01')),
                           (2, 2., np.datetime64('NaT')),
                           (3, 3., np.datetime64('2010-01-03'))],
                          dtype=[('a', 'i8'),
                                 ('b', 'f8'),
                                 ('date', 'datetime64[D]')]))

t = symbol('t', 'var * {a: int64, b: float64, date: ?date}')

to = symbol('to', 'var * {a: int64, b: float64}')
bo = bcolz.ctable(np.array([(1, 1.), (2, 2.), (3, np.nan)],
                           dtype=[('a', 'i8'), ('b', 'f8')]))


def test_discover():
    assert discover(b) == dshape('3 * {a: int64, b: float64, date: date}')
    assert discover(b['a']) == dshape('3 * int64')


def test_reductions():
    assert compute(t.a.sum(), b) == 6
    assert compute(t.a.min(), b) == 1
    assert compute(t.a.max(), b) == 3
    assert compute(t.a.mean(), b) == 2.0
    assert abs(compute(t.a.std(), b) - np.std([1, 2, 3])) < 1e-5
    assert abs(compute(t.a.var(), b) - np.var([1, 2, 3])) < 1e-5
    assert abs(compute(t.a.std(unbiased=True), b) - np.std([1, 2, 3],
                                                           ddof=1)) < 1e-5
    assert abs(compute(t.a.var(unbiased=True), b) - np.var([1, 2, 3],
                                                           ddof=1)) < 1e-5
    assert len(list(compute(t.distinct(), b))) == 3
    assert len(list(compute(t.a.distinct(), b))) == 3

    assert compute(t.a.nunique(), b) == 3
    assert isinstance(compute(t.a.nunique(), b), np.integer)

    assert compute(t.a.count(), b) == 3
    assert isinstance(compute(t.date.count(), b), np.integer)

    assert compute(t.date.nunique(), b) == 2
    assert isinstance(compute(t.date.nunique(), b), np.integer)

    assert compute(t.date.count(), b) == 2
    assert isinstance(compute(t.a.count(), b), np.integer)

    assert compute(t.a[0], b) == 1
    assert compute(t.a[-1], b) == 3
    assert compute(t[0], b) == compute(t[0], b)
    assert compute(t[-1], b) == compute(t[-1], b)


def test_nunique():
    assert compute(t.a.nunique(), b) == 3
    assert compute(t.nunique(), b) == 3


def test_selection_head():
    ds = dshape('var * {a: int32, b: int32, c: float64}')
    b = into(bcolz.ctable,
             [(i, i + 1, float(i) ** 2) for i in range(10)],
             dshape=ds)
    t = symbol('t', ds)

    # numpy reductions return numpy scalars
    assert compute((t.a < t.b).all(), b).item() is True
    assert list(compute(t[t.a < t.b].a.head(10), b)) == list(range(10))
    assert list(compute(t[t.a > t.b].a.head(10), b)) == []

    assert into([], compute(t[t.a + t.b > t.c], b)) == [(0, 1, 0),
                                                        (1, 2, 1),
                                                        (2, 3, 4)]
    assert len(compute(t[t.a + t.b > t.c].head(10), b))  # non-empty
    assert len(compute(t[t.a + t.b < t.c].head(10), b))  # non-empty


def test_selection_isnan():
    b = bcolz.ctable([[1, np.nan, 3], [1., 2., np.nan]], names=['a', 'b'])
    t = symbol('t', discover(b))
    lhs = compute(t[t.a.isnan()], b)
    rhs = np.array([(np.nan, 2.0)], dtype=b.dtype)

    for n in b.dtype.names:
        assert np.isclose(lhs[n], rhs[n], equal_nan=True).all()
        assert np.isclose(compute(t[~t.b.isnan()], b)[n],
                          np.array(
                              [(1, 1.0), (np.nan, 2.0)], dtype=b.dtype)[n],
                          equal_nan=True).all()


def test_count_isnan():
    assert compute(to.a[~to.b.isnan()].count(), bo) == 2


def test_count_isnan_struct():
    assert compute(t[~t.b.isnan()].count(), b) == 3


def test_nrows():
    assert compute(t.nrows, b) == len(b)


def test_nelements():
    assert compute(t.nelements(axis=0), b) == len(b)
    assert compute(t.nelements(), b) == len(b)


# This is no longer desired. Handled by compute_up
def dont_test_pre_compute():
    b = bcolz.ctable(np.array([(1, 1., 10.), (2, 2., 20.), (3, 3., 30.)],
                              dtype=[('a', 'i8'), ('b', 'f8'), ('c', 'f8')]))

    s = symbol('s', discover(b))

    result = pre_compute(s[['a', 'b']], b)
    assert result.names == ['a', 'b']


def eq(a, b):
    return np.array_equal(a, b)


def test_unicode_field_names():
    b = bcolz.ctable(np.array([(1, 1., 10.), (2, 2., 20.), (3, 3., 30.)],
                              dtype=[('a', 'i8'), ('b', 'f8'), ('c', 'f8')]))
    s = symbol('s', discover(b))

    assert eq(compute(s[u'a'], b)[:], compute(s['a'], b)[:])
    assert eq(compute(s[[u'a', u'c']], b)[:], compute(s[['a', 'c']], b)[:])
    assert eq(compute(s[u'a'], b)[:],
              compute(s['a'],  b)[:])
    assert eq(compute(s[[u'a', u'c']], b)[:],
              compute(s[['a', 'c']],  b)[:])


def test_chunksize_inference():
    b = bcolz.ctable(np.array([(1, 1., 10.), (2, 2., 20.), (3, 3., 30.)],
                              dtype=[('a', 'i8'), ('b', 'f8'), ('c', 'f8')]),
                     chunklen=2)
    assert get_chunksize(b) == 2


def test_notnull():
    with pytest.raises(AttributeError):
        t.b.notnull


def test_by_with_single_row():
    ct = bcolz.ctable([[1, 1, 3, 3], [1, 2, 3, 4]], names=list('ab'))
    t = symbol('t', discover(ct))
    subset = t[t.a == 3]
    expr = by(subset.a, b_sum=subset.b.sum())
    result = compute(expr, ct)
    expected = compute(expr, ct, optimize=False)
    tm.assert_frame_equal(result, expected)



from __future__ import absolute_import, division, print_function

from datashape.dispatch import namespace
from multipledispatch import dispatch
from functools import partial


__all__ = 'dispatch', 'namespace'


dispatch = partial(dispatch, namespace=namespace)




from __future__ import absolute_import, division, print_function

import numpy as np
import h5py
from multipledispatch import MDNotImplementedError
from datashape import DataShape, to_numpy
from toolz import curry

import psutil

from multiprocessing.pool import ThreadPool

from ..partition import partitions
from ..expr import Reduction, Field, symbol
from ..expr import Expr, Slice, ElemWise
from ..expr import nelements
from ..expr import path, shape, Symbol
from ..expr.split import split

from .core import compute, compute_single_object
from ..dispatch import dispatch
from ..utils import available_memory


__all__ = []


# h5py File and Group are mapping objects.
compute.register(Expr, (h5py.File, h5py.Group))(compute_single_object)


@dispatch(Slice, h5py.Dataset)
def pre_compute(expr, data, scope=None, **kwargs):
    """ Don't push slices into memory, they're about to come in anyway """
    return data


@dispatch(Expr, h5py.Dataset)
def pre_compute(expr, data, scope=None, **kwargs):
    """ Bring dataset into memory if it's small relative to memory """
    nbytes = data.size * data.dtype.alignment
    comfortable_memory = available_memory() / 4

    if nbytes < comfortable_memory:
        return data.value
    else:
        return data


@dispatch(Expr, h5py.Dataset)
def post_compute(expr, data, scope=None):
    """ Bring dataset into memory if it's small relative to memory """
    nbytes = data.size * data.dtype.alignment
    comfortable_memory = available_memory() / 4

    if nbytes < comfortable_memory:
        return data.value
    else:
        return data


@dispatch(Symbol, h5py.Dataset)
def optimize(expr, data):
    return expr


@dispatch(Expr, h5py.Dataset)
def optimize(expr, data):
    child = optimize(expr._inputs[0], data)
    if child is expr._inputs[0]:
        return expr
    else:
        return expr._subs({expr._inputs[0]: child})


@dispatch(Slice, (h5py.File, h5py.Group, h5py.Dataset))
def optimize(expr, data):
    child = expr._inputs[0]
    if (isinstance(child, ElemWise) and len(child._inputs) == 1
            and shape(child._inputs[0]) == shape(child)):
        grandchild = child._inputs[0][expr.index]
        grandchild = optimize(grandchild, data)
        return child._subs({child._inputs[0]: grandchild})
    if (isinstance(child, ElemWise) and len(child._inputs) == 2
        and shape(child) == shape(expr._inputs[0]) == shape(child._inputs[1])):
        lhs, rhs = child._inputs
        lhs = lhs[expr.index]
        rhs = rhs[expr.index]
        lhs = optimize(lhs, data)
        rhs = optimize(rhs, data)
        return child._subs(dict(zip(child._inputs, (lhs, rhs))))
    else:
        return expr


@dispatch(Symbol, (h5py.File, h5py.Group, h5py.Dataset))
def compute_up(expr, data, **kwargs):
    return data


@dispatch(Field, (h5py.File, h5py.Group))
def compute_up(expr, data, **kwargs):
    return data[expr._name]


@dispatch(Slice, h5py.Dataset)
def compute_up(expr, data, **kwargs):
    return data[expr.index]


@dispatch(nelements, h5py.Dataset)
def compute_up(expr, data, **kwargs):
    return compute_up.dispatch(type(expr), np.ndarray)(expr, data, **kwargs)


@dispatch(Expr, (h5py.File, h5py.Group))
def compute_down(expr, data, **kwargs):
    leaf = expr._leaves()[0]
    p = list(path(expr, leaf))[::-1][1:]
    if not p:
        return data
    for e in p:
        data = compute_up(e, data)
        if not isinstance(data, (h5py.File, h5py.Group)):
            break

    expr2 = expr._subs({e: symbol('leaf', e.dshape)})
    return compute_down(expr2, data, **kwargs)


def compute_chunk(source, target, chunk, chunk_expr, parts):
    """ Pull out a part, compute it, insert it into the target """
    source_part, target_part = parts
    part = source[source_part]
    result = compute(chunk_expr, {chunk: part}, return_type='native')
    target[target_part] = result


thread_pool = None

def _get_map(map):
    global thread_pool

    if map is None:
        if thread_pool is None:
            thread_pool = ThreadPool(psutil.cpu_count())
        return thread_pool.map
    else:
        return map


@dispatch(Expr, h5py.Dataset)
def compute_down(expr, data, map=None, **kwargs):
    """ Compute expressions on H5Py datasets by operating on chunks

    This uses blaze.expr.split to break a full-array-computation into a
    per-chunk computation and a on-aggregate computation.

    This uses blaze.partition to pick out chunks from the h5py dataset, uses
    compute(numpy) to compute on each chunk and then uses blaze.partition to
    aggregate these (hopefully smaller) intermediate results into a local
    numpy array.  It then performs a second operation (again given by
    blaze.expr.split) on this intermediate aggregate

    The expression must contain some sort of Reduction.  Both the intermediate
    result and the final result are assumed to fit into memory
    """
    map = _get_map(map)

    leaf = expr._leaves()[0]
    if not any(isinstance(node, Reduction) for node in path(expr, leaf)):
        raise MDNotImplementedError()

    # Compute chunksize (this should be improved)
    chunksize = kwargs.get('chunksize', data.chunks)

    # Split expression into per-chunk and on-aggregate pieces
    chunk = symbol('chunk', DataShape(*(chunksize + (leaf.dshape.measure,))))
    (chunk, chunk_expr), (agg, agg_expr) = \
            split(leaf, expr, chunk=chunk)

    # Create numpy array to hold intermediate aggregate
    shape, dtype = to_numpy(agg.dshape)
    intermediate = np.empty(shape=shape, dtype=dtype)

    # Compute partitions
    source_parts = list(partitions(data, chunksize=chunksize, keepdims=True))
    target_parts = list(partitions(intermediate, chunksize=chunk_expr.shape,
                                   keepdims=True))

    list(map(
        curry(compute_chunk, data, intermediate, chunk, chunk_expr),
        zip(source_parts, target_parts)
    ))

    # Compute on the aggregate
    return compute(agg_expr, {agg: intermediate}, return_type='native')



from .core import pre_compute
from ..dispatch import dispatch
from ..expr import Expr
from odo.backends.json import JSON, JSONLines
from odo import into
from collections import Iterator
from odo.utils import records_to_tuples


__all__ = ['pre_compute']


@dispatch(Expr, JSON)
def pre_compute(expr, data, **kwargs):
    seq = into(list, data, **kwargs)
    leaf = expr._leaves()[0]
    return list(records_to_tuples(leaf.dshape, seq))


@dispatch(Expr, JSONLines)
def pre_compute(expr, data, **kwargs):
    seq = into(Iterator, data, **kwargs)
    leaf = expr._leaves()[0]
    return records_to_tuples(leaf.dshape, seq)



from __future__ import absolute_import, division, print_function

from multipledispatch import MDNotImplementedError
from odo import Chunks, convert, into
from collections import Iterator, Iterable
from toolz import curry, concat
from datashape.dispatch import dispatch

import pandas as pd
import numpy as np

from ..expr import Head, ElemWise, Distinct, Symbol, Expr, path
from ..expr.split import split
from .core import compute
from .pmap import get_default_pmap


__all__ = ['Cheap', 'compute_chunk', 'compute_down']


Cheap = (Head, ElemWise, Distinct, Symbol)


@dispatch(Head, Chunks)
def pre_compute(expr, data, **kwargs):
    leaf = expr._leaves()[0]
    if all(isinstance(e, Cheap) for e in path(expr, leaf)):
        return convert(Iterator, data)
    else:
        raise MDNotImplementedError()


def compute_chunk(chunk, chunk_expr, part):
    return compute(chunk_expr, {chunk: part})


@dispatch(Expr, Chunks)
def compute_down(expr, data, map=None, **kwargs):
    if map is None:
        map = get_default_pmap()

    leaf = expr._leaves()[0]

    (chunk, chunk_expr), (agg, agg_expr) = split(leaf, expr)

    parts = list(map(curry(compute_chunk, chunk, chunk_expr), data))

    if isinstance(parts[0], np.ndarray):
        intermediate = np.concatenate(parts)
    elif isinstance(parts[0], pd.DataFrame):
        intermediate = pd.concat(parts)
    elif isinstance(parts[0], (Iterable, Iterator)):
        intermediate = list(concat(parts))

    return compute(agg_expr, {agg: intermediate})


Cheap = (Head, ElemWise, Distinct, Symbol)

@dispatch(Head, Chunks)
def compute_down(expr, data, **kwargs):
    leaf = expr._leaves()[0]
    if all(isinstance(e, Cheap) for e in path(expr, leaf)):
        return compute(expr, {leaf: into(Iterator, data)}, **kwargs)
    else:
        raise MDNotImplementedError()



from blaze.partition import *
from blaze.expr import shape

import numpy as np

x = np.arange(24).reshape(4, 6)


def eq(a, b):
    if isinstance(a == b, bool):
        return a == b
    if isinstance(a, np.ndarray) or isinstance(b, np.ndarray):
        return (a == b).all()
    else:
        return a == b


def test_partition_get():
    assert eq(partition_get(x, (0, slice(0, None)), chunksize=(1, 6)),
              x[0, :])
    assert eq(partition_get(x, (slice(0, None), 0), chunksize=(4, 1)),
              x[:, 0])
    assert eq(partition_get(x, (slice(2, 4), slice(0, 2)), chunksize=(2, 2)),
              x[2:4, 0:2])


def test_partition_set():
    x = np.arange(24).reshape(4, 6)
    partition_set(x,
                  (slice(0, 2), slice(0, 2)), np.array([[1, 1], [1, 1]]),
                  chunksize=(2, 2))
    assert (x[:2, :2] == 1).all()


def test_partition_set_1d():
    x = np.arange(24).reshape(4, 6)
    partition_set(x,
                  (slice(0, 4), 0), np.array([[1], [1], [1], [1]]),
                  chunksize=(4, 1),
                  keepdims=False)
    assert (x[:4, 0] == 1).all()


def test_partitions():
    assert list(partitions(x, chunksize=(1, 6))) == \
            [(i, slice(0, 6)) for i in range(4)]
    assert list(partitions(x, chunksize=(4, 1))) == \
            [(slice(0, 4), i) for i in range(6)]
    assert list(partitions(x, chunksize=(2, 3))) == [
            (slice(0, 2), slice(0, 3)), (slice(0, 2), slice(3, 6)),
            (slice(2, 4), slice(0, 3)), (slice(2, 4), slice(3, 6))]


def dont_test_partitions_flat():
    assert list(partitions(x, chunksize=(2, 3))) == [
            (slice(0, 2), slice(0, 3)), (slice(0, 2), slice(3, 6)),
            (slice(2, 4), slice(0, 3)), (slice(2, 4), slice(3, 6))]


def test_uneven_partitions():
    x = np.arange(10*12).reshape(10, 12)
    parts = list(partitions(x, chunksize=(7, 7)))

    assert len(parts) == 2 * 2

    assert parts == [(slice(0,  7), slice(0, 7)), (slice(0,  7), slice(7, 12)),
                     (slice(7, 10), slice(0, 7)), (slice(7, 10), slice(7, 12))]

    x = np.arange(20*24).reshape(20, 24)
    parts = list(partitions(x, chunksize=(7, 7)))


def test_3d_partitions():
    x = np.arange(4*4*6).reshape(4, 4, 6)
    parts = list(partitions(x, chunksize=(2, 2, 3)))
    assert len(parts) == 2 * 2 * 2



import pytest

from datashape import dshape
from datashape.util.testing import assert_dshape_equal

from blaze.expr import symbol
from blaze.expr.core import common_subexpression
from blaze.expr.collections import merge, join, transform, concat
from blaze.utils import raises
from blaze.compatibility import builtins
from toolz import isdistinct


t = symbol('t', '5 * {name: string, amount: int, x: real}')


def test_merge():
    e = symbol('e', '3 * 5 * {name: string, amount: int, x: real}')
    expr = merge(name=e.name, y=e.x)

    assert set(expr.fields) == set(['name', 'y'])
    assert expr.y.isidentical(e.x.label('y'))


@pytest.mark.parametrize('bunk_column',
                         ['deadbeef',
                          ['deadbeef'],
                          ['name', 'deadbeef'],
                          10,
                          [10],
                          [None]])
def test_sort_validation_tabular(bunk_column):
    t = symbol('t', 'var * {name: string, id: int32}')
    with pytest.raises(ValueError):
        t.sort(bunk_column)


@pytest.mark.xfail(reason=("Can't validate sort key "
                           "when sorting array-like dataset."))
def test_sort_validation_array():
    t = symbol('t', '10 * 20 * int32')
    with pytest.raises(ValueError):
        t.sort('foobar')


def test_sort_validation_ascending():
    t = symbol('t', '10 * 20 * int32')
    with pytest.raises(ValueError):
        t.sort(ascending=None)
    with pytest.raises(ValueError):
        t.sort(ascending='zzz')


def test_merge_options():
    s = symbol('s', 'var * {a: ?A, b: ?B}')

    merged = merge(a=s.a, b=s.b)
    assert_dshape_equal(merged.dshape, dshape('var * {a: ?A, b: ?B}'))
    assert_dshape_equal(merged.a.dshape, dshape('var * ?A'))
    assert_dshape_equal(merged.b.dshape, dshape('var * ?B'))


def test_merge_on_single_argument_is_noop():
    assert merge(t.name).isidentical(t.name)


def test_merge_common_subexpression():
    t = symbol('t', 'var * {a: float64}')
    result = common_subexpression(t.a - t.a % 3, t.a % 3)
    assert result.isidentical(t.a)


def test_merge_exceptions():
    t = symbol('t', 'var * {x:int}')
    with pytest.raises(ValueError) as excinfo:
        merge(t, x=2*t.x)

    assert "Repeated columns" in str(excinfo.value)


def test_transform():
    t = symbol('t', '5 * {x: int, y: int}')
    expr = transform(t, z=t.x + t.y)
    assert expr.fields == ['x', 'y', 'z']

    assert builtins.any(
        (t.x + t.y).isidentical(node) for node in expr._subterms()
    )


def test_distinct():
    assert '5' not in str(t.distinct().dshape)


def test_distinct_exceptions():
    t = symbol('t', 'var * {x:int}')
    with pytest.raises(ValueError) as excinfo:
        t.distinct('a')

    assert "a is not a field of t" == str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        t.distinct(10)

    assert "on must be a name or field, not: 10" == str(excinfo.value)

    s = symbol('s', 'var * {y:int}')
    with pytest.raises(ValueError) as excinfo:
        t.distinct(s.y)

    assert "s.y is not a field of t" == str(excinfo.value)


def test_sample_exceptions():
    t = symbol('t', 'var * {x:int, y:int}')
    with pytest.raises(ValueError):
        t.sample(n=1, frac=0.1)
    with pytest.raises(TypeError):
        t.sample(n='a')
    with pytest.raises(ValueError):
        t.sample(frac='a')
    with pytest.raises(TypeError):
        t.sample(foo='a')
    with pytest.raises(TypeError):
        t.sample()


def test_join_on_same_columns():
    a = symbol('a', 'var * {x: int, y: int, z: int}')
    b = symbol('b', 'var * {x: int, y: int, w: int}')

    c = join(a, b, 'x')

    assert isdistinct(c.fields)
    assert len(c.fields) == 5
    assert 'y_left' in c.fields
    assert 'y_right' in c.fields


def test_join_on_same_table():
    a = symbol('a', 'var * {x: int, y: int}')

    c = join(a, a, 'x')

    assert isdistinct(c.fields)
    assert len(c.fields) == 3


def test_join_suffixes():
    a = symbol('a', 'var * {x: int, y: int}')
    b = join(a, a, 'x', suffixes=('_l', '_r'))

    assert isdistinct(b.fields)
    assert len(b.fields) == 3
    assert set(b.fields) == set(['x', 'y_l', 'y_r'])


def test_join_on_single_column():
    a = symbol('a', 'var * {x: int, y: int, z: int}')
    b = symbol('b', 'var * {x: int, y: int, w: int}')

    expr = join(a, b.x)

    assert expr.on_right == 'x'


def test_raise_error_if_join_on_no_columns():
    a = symbol('a', 'var * {x: int}')
    b = symbol('b', 'var * {y: int}')

    assert raises(ValueError, lambda: join(a, b))


def test_join_option_types():
    a = symbol('a', 'var * {x: ?int}')
    b = symbol('b', 'var * {x: int}')

    assert join(a, b, 'x').dshape == dshape('var * {x: int}')
    assert join(b, a, 'x').dshape == dshape('var * {x: int}')


@pytest.mark.xfail
def test_join_option_types_outer():
    a = symbol('a', 'var * {x: ?int}')
    b = symbol('b', 'var * {x: int}')

    assert (join(a, b, 'x', how='outer').dshape ==
            join(b, a, 'x', how='outer').dshape ==
            dshape('var * {x: ?int}'))


def test_join_option_string_types():
    a = symbol('a', 'var * {x: ?string}')
    b = symbol('b', 'var * {x: string}')
    c = symbol('c', 'var * {x: ?string}')

    assert (join(a, b, 'x').dshape ==
            join(b, a, 'x').dshape ==
            dshape('var * {x: string}'))

    assert (join(a, c, 'x').dshape ==
            join(c, a, 'x').dshape ==
            dshape('var * {x: ?string}'))


def test_join_exceptions():
    """
    exception raised for mismatched schema;
    exception raised for no shared fields
    """
    a = symbol('a', 'var * {x: int}')
    b = symbol('b', 'var * {x: string}')
    with pytest.raises(TypeError) as excinfo:
        join(a, b, 'x')

    assert "Schemata of joining columns do not match," in str(excinfo.value)
    assert "x=int32 and x=string" in str(excinfo.value)

    b = symbol('b', 'var * {z: int}')
    with pytest.raises(ValueError) as excinfo:
        join(a, b)

    assert "No shared columns between a and b" in str(excinfo.value)

    b = symbol('b', 'var * {x: int}')
    with pytest.raises(ValueError) as excinfo:
        join(a, b, how='inner_')

    assert "Got: inner_" in str(excinfo.value)


def test_join_type_promotion():
    a = symbol('a', 'var * {x: int32}')
    b = symbol('b', 'var * {x: int64}')

    assert join(a, b, 'x').dshape == dshape('var * {x: int64}')


def test_join_type_promotion_option():
    a = symbol('a', 'var * {x: ?int32}')
    b = symbol('b', 'var * {x: int64}')

    assert join(a, b, 'x').dshape == dshape('var * {x: int64}')


def test_isin():
    a = symbol('a', 'var * {x: int, y: string}')
    assert hasattr(a.x, 'isin')
    assert hasattr(a.y, 'isin')
    assert not hasattr(a, 'isin')


def test_isin_no_expressions():
    a = symbol('a', 'var * int')
    b = symbol('b', 'var * int')
    with pytest.raises(TypeError):
        a.isin(b)


def test_concat_table():
    a = symbol('a', '3 * {a: int32, b: int32}')
    b = symbol('a', '5 * {a: int32, b: int32}')
    v = symbol('v', 'var * {a: int32, b: int32}')

    assert concat(a, b).dshape == dshape('8 * {a: int32, b: int32}')
    assert concat(a, v).dshape == dshape('var * {a: int32, b: int32}')


def test_concat_mat():
    a = symbol('a', '3 * 5 * int32')
    b = symbol('b', '3 * 5 * int32')
    v = symbol('v', 'var * 5 * int32')
    u = symbol('u', '3 * var * int32')

    assert concat(a, b, axis=0).dshape == dshape('6 * 5 * int32')
    assert concat(a, b, axis=1).dshape == dshape('3 * 10 * int32')
    assert concat(a, v, axis=0).dshape == dshape('var * 5 * int32')
    assert concat(a, u, axis=1).dshape == dshape('3 * var * int32')


def test_concat_arr():
    a = symbol('a', '3 * int32')
    b = symbol('b', '5 * int32')
    v = symbol('v', 'var * int32')

    assert concat(a, b).dshape == dshape('8 * int32')
    assert concat(a, v).dshape == dshape('var * int32')


def test_concat_different_measure():
    a = symbol('a', '3 * 5 * int32')
    b = symbol('b', '3 * 5 * float64')

    with pytest.raises(TypeError) as excinfo:
        concat(a, b)

    msg = 'Mismatched measures: {l} != {r}'.format(l=a.dshape.measure,
                                                   r=b.dshape.measure)
    assert msg == str(excinfo.value)


def test_concat_different_along_concat_axis():
    a = symbol('a', '3 * 5 * int32')
    b = symbol('b', '3 * 6 * int32')

    with pytest.raises(TypeError) as excinfo:
        concat(a, b, axis=0)

    assert "not equal along axis 1: 5 != 6" in str(excinfo.value)

    b = symbol('b', '4 * 6 * int32')
    with pytest.raises(TypeError) as excinfo:
        concat(a, b, axis=1)

    assert "not equal along axis 0: 3 != 4" in str(excinfo.value)


def test_concat_negative_axis():
    a = symbol('a', '3 * 5 * int32')
    b = symbol('b', '3 * 5 * int32')

    with pytest.raises(ValueError) as excinfo:
        concat(a, b, axis=-1)

    assert "must be in range: [0, 2)" in str(excinfo.value)


def test_concat_axis_too_great():
    a = symbol('a', '3 * 5 * int32')
    b = symbol('b', '3 * 5 * int32')

    with pytest.raises(ValueError) as excinfo:
        concat(a, b, axis=2)

    assert "must be in range: [0, 2)" in str(excinfo.value)


def test_shift():
    t = symbol('t', 'var * float64')
    assert t.shift(1).dshape == dshape('var * ?float64')
    assert t.shift(0).dshape == t.dshape
    assert t.shift(-1).dshape == dshape('var * ?float64')
    assert str(t.shift(1)) == 'shift(t, n=1)'
    assert str(t.shift(0)) == 'shift(t, n=0)'
    assert str(t.shift(-1)) == 'shift(t, n=-1)'


def test_shift_not_int():
    a = symbol('a', '3 * {x: int32, y: int32}')
    with pytest.raises(TypeError):
        a.x.shift(1.3)



from blaze.compute.pydatetime import truncate
from datetime import datetime, date, timedelta

def test_hour():
    dts = [datetime(2000, 6, 20,  1, 00, 00),
           datetime(2000, 6, 20, 12, 59, 59),
           datetime(2000, 6, 20, 12, 00, 00),
           datetime(2000, 6, 20, 11, 59, 59)]

    assert [truncate(dt, 1, 'hour') for dt in dts] == \
            [datetime(2000, 6, 20,  1, 0),
             datetime(2000, 6, 20, 12, 0),
             datetime(2000, 6, 20, 12, 0),
             datetime(2000, 6, 20, 11, 0)]


def test_month():
    dts = [datetime(2000, 7, 1),
           datetime(2000, 6, 30),
           datetime(2000, 6, 1),
           datetime(2000, 5, 31)]

    assert [truncate(dt, 1, 'month') for dt in dts] == \
            [date(2000, 7, 1),
             date(2000, 6, 1),
             date(2000, 6, 1),
             date(2000, 5, 1),]

    assert truncate(datetime(2000, 12, 1), 1, 'month') == \
            date(2000, 12, 1)


def test_week():
    d = date(2014, 11, 8)
    assert truncate(d, 1, 'week').isoweekday() == 7
    assert (d - truncate(d, 1, 'week')) < timedelta(days=7)
    assert (d - truncate(d, 1, 'week')) > timedelta(days=0)



from __future__ import absolute_import, division, print_function

from blaze.expr.functions import *
from blaze.expr import Expr, symbol
from blaze.expr import math
import numpy as np


def test_reductions():
    assert max([1, 2, 3]) == 3
    assert max(np.array([1, 2, 3])) == 3
    assert isinstance(max(symbol('t', 'var * {x: int, y: int}').x), Expr)
    assert all([True, True, True]) is True
    assert all([True, True, False]) is False


def test_math():
    assert sin(0) == 0
    assert isinstance(sin(0), float)
    assert isinstance(sin(np.int32(0)), np.float)
    assert isinstance(sin(symbol('x', 'real')), Expr)


def test_abs():
    assert abs(-1) == 1
    assert (abs(np.array([-1, 2, 3])) == np.array([1, 2, 3])).all()
    x = symbol('x', 'int')
    assert isinstance(abs(x), math.abs)



from __future__ import absolute_import, division, print_function

from datashape import Option, real, int_, bool_, isreal, isnumeric

from .arithmetic import UnaryOp, BinOp, Arithmetic
from .expressions import schema_method_list
from ..compatibility import builtins


# Here follows a large number of unary operators.  These were selected by
# taking the intersection of the functions in ``math`` and ``numpy``

__all__ = ['abs', 'sqrt', 'sin', 'sinh', 'cos', 'cosh', 'tan', 'tanh', 'exp',
           'expm1', 'log', 'log10', 'log1p', 'acos', 'acosh', 'asin', 'asinh',
           'atan', 'atanh', 'radians', 'degrees', 'atan2', 'ceil', 'floor',
           'trunc', 'isnan', 'notnull', 'UnaryMath', 'BinaryMath',
           'greatest', 'least']


class UnaryMath(UnaryOp):

    """Mathematical unary operator with real valued dshape like sin, or exp
    """
    _arguments = '_child',
    _dtype = real


class BinaryMath(BinOp):
    _dtype = real

    def __str__(self):
        return '%s(%s, %s)' % (type(self).__name__, self.lhs, self.rhs)


_unary_math_names = (
    'abs',
    'sqrt',
    'sin',
    'sinh',
    'cos',
    'cosh',
    'tan',
    'tanh',
    'exp',
    'expm1',
    'log',
    'log10',
    'log1p',
    'acos',
    'acosh',
    'asin',
    'asinh',
    'atan',
    'atanh',
    'radians',
    'degrees',
)


for name in _unary_math_names:
    locals()[name] = type(name, (UnaryMath,), {})


_binary_math_names = (
    'atan2',
    'copysign',
    'fmod',
    'hypot',
    'ldexp',
)

for name in _binary_math_names:
    locals()[name] = type(name, (BinaryMath,), {})


class greatest(Arithmetic):
    _arguments = 'lhs', 'rhs'
    op = builtins.max

    def __str__(self):
        return 'greatest(%s, %s)' % (self.lhs, self.rhs)


class least(Arithmetic):
    _arguments = 'lhs', 'rhs'
    op = builtins.min

    def __str__(self):
        return 'least(%s, %s)' % (self.lhs, self.rhs)


_unary_integer_math = (
    'ceil',
    'floor',
    'trunc',
)

for name in _unary_integer_math:
    locals()[name] = type(name, (UnaryMath,), dict(_dtype=int_))


class isnan(UnaryMath):
    _dtype = bool_


class notnull(UnaryOp):

    """ Return whether an expression is not null

    Examples
    --------
    >>> from blaze import symbol, compute
    >>> s = symbol('s', 'var * int64')
    >>> expr = notnull(s)
    >>> expr.dshape
    dshape("var * bool")
    >>> list(compute(expr, [1, 2, None, 3]))
    [True, True, False, True]
    """
    _dtype = bool_


def truncate(expr, precision):
    """ Truncate number to precision

    Examples
    --------
    >>> from blaze import symbol, compute
    >>> x = symbol('x', 'real')
    >>> compute(x.truncate(10), 123)
    120
    >>> compute(x.truncate(0.1), 3.1415)  # doctest: +SKIP
    3.1
    """
    return expr // precision * precision


schema_method_list.extend([
    (isreal, set([isnan])),
    (isnumeric, set([truncate])),
    (lambda ds: isinstance(ds, Option) or isinstance(ds.measure, Option),
     set([notnull]))
])



from datashape import dshape
from datashape.util.testing import assert_dshape_equal
import pytest

from blaze import symbol
import blaze as bz
import blaze.expr.strings as bzs

dshapes = ['var * {name: string}',
           'var * {name: ?string}',
           'var * string',
           'var * ?string',
           'string']

lhsrhs_ds = ['var * {name: string, comment: string[25]}',
             'var * {name: string[10], comment: string}',
             'var * {name: string, comment: string}',
             'var * {name: ?string, comment: string}',
             'var * {name: string, comment: ?string}',
             '10 * {name: string, comment: ?string}']


@pytest.fixture(scope='module')
def strcat_sym():
    '''
    blaze symbol used to test exceptions raised by cat()
    '''
    ds = dshape('3 * {name: string, comment: string, num: int32}')
    s = symbol('s', dshape=ds)
    return s


@pytest.mark.parametrize('ds', dshapes)
def test_like(ds):
    t = symbol('t', ds)
    expr = getattr(t, 'name', t).str.like('Alice*')
    assert expr.pattern == 'Alice*'
    assert_dshape_equal(
        expr.schema.measure,
        dshape('%sbool' % ('?' if '?' in ds else '')).measure,
    )


@pytest.mark.parametrize('ds', dshapes)
def test_upper_schema(ds):
    t = symbol('t', ds)
    expr_upper = getattr(t, 'name', t).str.upper()
    expr_lower = getattr(t, 'name', t).str.lower()
    assert (expr_upper.schema.measure ==
            expr_lower.schema.measure ==
            dshape('%sstring' % ('?' if '?' in ds else '')).measure)


def test_str_namespace():
    t = symbol('t', 'var * {name: string}')
    assert bzs.upper(t.name).isidentical(t.name.str.upper())
    assert bzs.lower(t.name).isidentical(t.name.str.lower())
    assert (bzs.lower(bzs.upper(t.name))
            .isidentical(t.name.str.upper().str.lower()))
    assert bzs.len(t.name).isidentical(t.name.str.len())
    assert bzs.like(t.name, '*a').isidentical(t.name.str.like('*a'))
    assert (bzs.cat(bzs.cat(t.name, t.name, sep=' ++ '), t.name)
            .isidentical(t.name.str.cat(t.name, sep=' ++ ')
                               .str.cat(t.name)))
    assert bzs.isalnum(t.name).isidentical(t.name.str.isalnum())
    assert bzs.isalpha(t.name).isidentical(t.name.str.isalpha())
    assert bzs.isdecimal(t.name).isidentical(t.name.str.isdecimal())
    assert bzs.isdigit(t.name).isidentical(t.name.str.isdigit())
    assert bzs.islower(t.name).isidentical(t.name.str.islower())
    assert bzs.isnumeric(t.name).isidentical(t.name.str.isnumeric())
    assert bzs.isspace(t.name).isidentical(t.name.str.isspace())
    assert bzs.istitle(t.name).isidentical(t.name.str.istitle())
    assert bzs.isupper(t.name).isidentical(t.name.str.isupper())

    assert bzs.replace(t.name, 'A', 'a').isidentical(t.name.str.replace('A', 'a'))
    assert bzs.capitalize(t.name).isidentical(t.name.str.capitalize())
    assert bzs.strip(t.name).isidentical(t.name.str.strip())
    assert bzs.lstrip(t.name).isidentical(t.name.str.lstrip())
    assert bzs.rstrip(t.name).isidentical(t.name.str.rstrip())
    assert bzs.pad(t.name, 5).isidentical(t.name.str.pad(5))
    assert (bzs.slice_replace(t.name, 1, 3, 'foo')
            .isidentical(t.name.str.slice_replace(1, 3, 'foo')))

@pytest.mark.parametrize('ds', lhsrhs_ds)
def test_cat_schema_shape(ds):
    t = symbol('t', ds)
    expr = t.name.str.cat(t.comment)
    assert (expr.schema.measure ==
            dshape('%sstring' % ('?' if '?' in ds else '')).measure)
    assert expr.lhs.shape == expr.rhs.shape == expr.shape


def test_cat_exception_non_string_sep(strcat_sym):
    with pytest.raises(TypeError):
        strcat_sym.name.str.cat(strcat_sym.comment, sep=123)


def test_cat_exception_non_string_col_to_cat(strcat_sym):
    with pytest.raises(TypeError):
        strcat_sym.name.str.cat(strcat_sym.num)




from __future__ import absolute_import, division, print_function

import pytest
import os
import json
from datetime import datetime, timedelta

import pandas as pd
from pytz import utc

from blaze.server.serialization import (json_dumps, json_dumps_trusted,
                                        object_hook, object_hook_trusted)


@pytest.mark.parametrize('serializers', [(json_dumps, object_hook),
                                          (json_dumps_trusted, object_hook_trusted)])
@pytest.mark.parametrize('input_,serialized', (
    ([1, datetime(2000, 1, 1, 12, 30, 0, 0, utc)],
     '[1, {"__!datetime": "2000-01-01T12:30:00+00:00"}]'),
    ([1, pd.NaT], '[1, {"__!datetime": "NaT"}]'),
    ([1, frozenset([1, 2, 3])], '[1, {"__!frozenset": [1, 2, 3]}]'),
    ([1, timedelta(seconds=5)], '[1, {"__!timedelta": 5.0}]'),
))
def test_json_encoder(serializers, input_, serialized):
    serializer, deserializer = serializers
    result = json.dumps(input_, default=serializer)
    assert result == serialized
    assert json.loads(result, object_hook=deserializer) == input_



import pytest
tables = pytest.importorskip('tables')

from blaze.utils import tmpfile
from blaze import symbol, discover, compute, pre_compute
import pandas as pd
from datetime import datetime
from odo import Chunks, odo


@pytest.fixture
def df():
    return pd.DataFrame([['a', 1, 10., datetime(2000, 1, 1)],
                         ['ab', 2, 20., datetime(2000, 2, 2)],
                         ['abc', 3, 30., datetime(2000, 3, 3)],
                         ['abcd', 4, 40., datetime(2000, 4, 4)]],
                        columns=['name', 'a', 'b', 'time'])


@pytest.fixture
def s(hdf):
    return symbol('s', discover(hdf))


@pytest.yield_fixture(params=['fixed', 'table'])
def hdf(df, request):
    with tmpfile('.hdf5') as fn:
        df.to_hdf(fn, '/data', format=request.param)
        df.to_hdf(fn, '/nested/data', format=request.param)
        with pd.HDFStore(fn, mode='r') as r:
            yield r


def test_basic_compute(hdf, s):
    result = compute(s.data, hdf)
    types = (
        pd.DataFrame,
        pd.io.pytables.Fixed,
        pd.io.pytables.AppendableFrameTable,
        Chunks
    )
    assert isinstance(result, types)


def test_pre_compute(hdf, s):
    result = pre_compute(s, hdf.get_storer('data'))
    assert isinstance(result, (pd.DataFrame, Chunks))


def test_groups(hdf, df, s):
    assert discover(hdf) == discover(dict(data=df, nested=dict(data=df)))
    assert odo(compute(s.nested.data.a, hdf), list) == [1, 2, 3, 4]



""" Python compute layer

>>> from blaze import symbol, compute

>>> accounts = symbol('accounts', 'var * {name: string, amount: int}')
>>> deadbeats = accounts[accounts['amount'] < 0]['name']

>>> data = [['Alice', 100], ['Bob', -50], ['Charlie', -20]]
>>> list(compute(deadbeats, data))
['Bob', 'Charlie']
"""
from __future__ import absolute_import, division, print_function

from collections import Mapping
import itertools
import numbers
import fnmatch
import operator
import datetime
import math
import random

from collections import Iterator
from functools import partial

from datashape import Option, to_numpy_dtype
import numpy as np
import pandas as pd
import toolz
from toolz import map, filter, compose, juxt, identity, tail

try:
    from cytoolz import groupby, reduceby, unique, take, concat, nth, pluck
except ImportError:
    from toolz import groupby, reduceby, unique, take, concat, nth, pluck

from datashape.predicates import isscalar, iscollection

from ..dispatch import dispatch
from ..expr import (Projection, Field, Broadcast, Map, Label, ReLabel,
                    Merge, Join, Selection, Reduction, Distinct,
                    By, Sort, Head, Sample, Apply, Summary, Like, IsIn,
                    DateTime, Date, Time, Millisecond, ElemWise,
                    Symbol, Slice, Expr, Arithmetic, ndim, DateTimeTruncate,
                    UTCFromTimestamp, notnull, UnaryMath, greatest, least,
                    Coerce)
from ..expr import reductions
from ..expr import count, nunique, mean, var, std
from ..expr import BinOp, UnaryOp, USub, Not, nelements
from ..compatibility import builtins, apply, unicode, _inttypes
from .core import compute, compute_up, optimize, base

from ..utils import listpack
from ..expr.broadcast import broadcast_collect
from .pyfunc import lambdify, funcstr
from . import pydatetime

# Dump exp, log, sin, ... into namespace
from math import (
    atan2,
    atanh,
    ceil,
    copysign,
    cos,
    cosh,
    degrees,
    e,
    erf,
    erfc,
    exp,
    expm1,
    fabs,
    factorial,
    floor,
    fmod,
    frexp,
    fsum,
    gamma,
    hypot,
    isinf,
    isnan,
    ldexp,
    lgamma,
    log,
    log10,
    log1p,
    modf,
    pi,
    pow,
    radians,
    sin,
    sinh,
    sqrt,
    tan,
    tanh,
    trunc,
)



__all__ = ['compute', 'compute_up', 'Sequence', 'rowfunc', 'rrowfunc']

Sequence = (tuple, list, Iterator, type(dict().items()))


@dispatch(Expr, Sequence)
def pre_compute(expr, seq, scope=None, **kwargs):
    try:
        if isinstance(seq, Iterator):
            first = next(seq)
            seq = concat([[first], seq])
        else:
            first = next(iter(seq))
    except StopIteration:
        return []
    if isinstance(first, Mapping):
        leaf = expr._leaves()[0]
        return pluck(leaf.fields, seq)
    else:
        return seq


@dispatch(Expr, Sequence)
def optimize(expr, seq):
    return broadcast_collect(expr)


def child(x):
    if hasattr(x, '_child'):
        return x._child
    if hasattr(x, '_inputs') and len(x._inputs) == 1:
        return x._inputs[0]
    raise NotImplementedError("Found expression with multiple children.\n"
                              "Perhaps Broadcast Optimize was not called")


def recursive_rowfunc(t, stop):
    """ Compose rowfunc functions up a tree

    >>> from blaze import symbol
    >>> accounts = symbol('accounts', 'var * {name: string, amount: int}')
    >>> expr = accounts['amount'].map(lambda x: x + 1)
    >>> f = recursive_rowfunc(expr, accounts)

    >>> row = ('Alice', 100)
    >>> f(row)
    101

    """
    funcs = []
    while not t.isidentical(stop):
        funcs.append(rowfunc(t))
        t = child(t)
    return compose(*funcs)


rrowfunc = recursive_rowfunc


@dispatch(Symbol)
def rowfunc(t):
    return identity


@dispatch(Projection)
def rowfunc(t):
    """ Rowfunc provides a function that can be mapped onto a sequence.

    >>> from blaze import symbol
    >>> accounts = symbol('accounts', 'var * {name: string, amount: int}')
    >>> f = rowfunc(accounts['amount'])

    >>> row = ('Alice', 100)
    >>> f(row)
    100

    See Also:
        compute<Rowwise, Sequence>
    """
    from toolz.itertoolz import getter
    indices = [t._child.fields.index(col) for col in t.fields]
    return getter(indices)


@dispatch(Field)
def rowfunc(t):
    index = t._child.fields.index(t._name)
    return lambda x, index=index: x[index]


@dispatch(IsIn)
def rowfunc(t):
    return t._keys.__contains__


@dispatch(Broadcast)
def rowfunc(t):
    return lambdify(t._scalars, t._scalar_expr)


@dispatch(Arithmetic)
def rowfunc(expr):
    return eval(funcstr(expr))


@dispatch(Map)
def rowfunc(t):
    if isscalar(t._child.dshape.measure):
        return t.func
    else:
        return partial(apply, t.func)


@dispatch((Label, ReLabel))
def rowfunc(t):
    return identity


@dispatch(DateTime)
def rowfunc(t):
    return lambda row: getattr(row, t.attr)


@dispatch(UTCFromTimestamp)
def rowfunc(t):
    return datetime.datetime.utcfromtimestamp


@dispatch((Date, Time))
def rowfunc(t):
    return lambda row: getattr(row, t.attr)()


@dispatch(Millisecond)
def rowfunc(_):
    return lambda row: getattr(row, 'microsecond') // 1000


@dispatch(DateTimeTruncate)
def rowfunc(expr):
    return partial(pydatetime.truncate, measure=expr.measure, unit=expr.unit)


@dispatch(UnaryMath)
def rowfunc(expr):
    return getattr(math, type(expr).__name__)


@dispatch(USub)
def rowfunc(expr):
    return operator.neg


@dispatch(Not)
def rowfunc(expr):
    return operator.invert


@dispatch(Arithmetic)
def rowfunc(expr):
    if not isinstance(expr.lhs, Expr):
        return lambda x: expr.op(expr.lhs, x)
    if not isinstance(expr.rhs, Expr):
        return lambda x: expr.op(x, expr.rhs)
    return expr.op


@dispatch(ElemWise, base)
def compute_up(expr, data, **kwargs):
    return rowfunc(expr)(data)


@dispatch(notnull)
def rowfunc(_):
    return lambda x: x is not None


def concat_maybe_tuples(vals):
    """

    >>> concat_maybe_tuples([1, (2, 3)])
    (1, 2, 3)
    """
    result = []
    for v in vals:
        if isinstance(v, (tuple, list)):
            result.extend(v)
        else:
            result.append(v)
    return tuple(result)


def deepmap(func, *data, **kwargs):
    """

    >>> inc = lambda x: x + 1
    >>> list(deepmap(inc, [1, 2], n=1))
    [2, 3]
    >>> list(deepmap(inc, [(1, 2), (3, 4)], n=2))
    [(2, 3), (4, 5)]

    Works on variadic args too

    >>> add = lambda x, y: x + y
    >>> list(deepmap(add, [1, 2], [10, 20], n=1))
    [11, 22]
    """
    n = kwargs.pop('n', 1)
    if n == 0:
        return func(*data)
    if n == 1:
        return map(func, *data)
    else:
        return map(compose(tuple, partial(deepmap, func, n=n - 1)), *data)


@dispatch(Merge)
def rowfunc(t):
    children = [optimize(child, []) for child in t.children]
    funcs = [rrowfunc(_child, t._child) for _child in children]
    return compose(concat_maybe_tuples, juxt(*funcs))


@dispatch(ElemWise, Sequence)
def compute_up(t, seq, **kwargs):
    func = rowfunc(t)
    if iscollection(t._child.dshape):
        return deepmap(func, seq, n=ndim(child(t)))
    else:
        return func(seq)


@dispatch(Broadcast, Sequence)
def compute_up(t, seq, **kwargs):
    func = rowfunc(t)
    return deepmap(func, seq, n=ndim(child(t)))


@dispatch(Arithmetic, Sequence + (int, float, bool))
def compute_up(t, seq, **kwargs):
    func = rowfunc(t)
    return deepmap(func, seq, n=ndim(child(t)))


@dispatch(Arithmetic, Sequence, Sequence)
def compute_up(t, a, b, **kwargs):
    if ndim(t.lhs) != ndim(t.rhs):
        raise ValueError()
    # TODO: Tee if necessary
    func = rowfunc(t)
    return deepmap(func, a, b, n=ndim(t.lhs))


@dispatch(Selection, Sequence)
def compute_up(t, seq, **kwargs):
    predicate = optimize(t.predicate, seq)
    predicate = rrowfunc(predicate, child(t))
    return filter(predicate, seq)


@dispatch(Selection, Sequence, Sequence)
def compute_up(expr, seq, predicate, **kwargs):
    preds = iter(predicate)
    return filter(lambda _: next(preds), seq)


@dispatch(Reduction, Sequence)
def compute_up(t, seq, **kwargs):
    if t.axis != (0,):
        raise NotImplementedError('Only 1D reductions currently supported')
    result = compute_up_1d(t, seq, **kwargs)
    if t.keepdims:
        return (result,)
    else:
        return result


@dispatch(Reduction, Sequence)
def compute_up_1d(t, seq, **kwargs):
    op = getattr(builtins, t.symbol)
    return op(seq)


@dispatch(nelements, Sequence)
def compute_up_1d(expr, seq, **kwargs):
    try:
        return len(seq)
    except TypeError:
        return toolz.count(seq)


@dispatch(ElemWise, base)
def compute_up(expr, data, **kwargs):
    return rowfunc(expr)(data)


@dispatch(BinOp, base, base)
def compute_up(bop, a, b, **kwargs):
    return bop.op(a, b)


@dispatch(UnaryOp, base)
def compute_up(uop, x, **kwargs):
    return uop.op(x)


@dispatch(UnaryMath, numbers.Real)
def compute_up(f, n, **kwargs):
    return getattr(math, type(f).__name__)(n)


def _mean(seq):
    total = 0
    count = 0
    for item in seq:
        total += item
        count += 1
    return float(total) / count


def _var(seq, unbiased):
    total = 0
    total_squared = 0
    count = 0
    for item in seq:
        total += item
        total_squared += item * item
        count += 1

    return (total_squared - (total * total) / count) / (count - unbiased)


def _std(seq, unbiased):
    return math.sqrt(_var(seq, unbiased))


@dispatch(count, Sequence)
def compute_up_1d(t, seq, **kwargs):
    return toolz.count(filter(None, seq))


@dispatch(Distinct, Sequence)
def compute_up(t, seq, **kwargs):
    if t.on:
        raise NotImplementedError(
            'python backend cannot specify what columns to distinct on'
        )
    try:
        row = toolz.first(seq)
    except StopIteration:
        return ()
    seq = concat([[row], seq])  # re-add row to seq

    if isinstance(row, list):
        seq = map(tuple, seq)

    return unique(seq)


@dispatch(nunique, Sequence)
def compute_up_1d(t, seq, **kwargs):
    return len(set(seq))


@dispatch(mean, Sequence)
def compute_up_1d(t, seq, **kwargs):
    return _mean(seq)


@dispatch(var, Sequence)
def compute_up_1d(t, seq, **kwargs):
    return _var(seq, t.unbiased)


@dispatch(std, Sequence)
def compute_up_1d(t, seq, **kwargs):
    return _std(seq, t.unbiased)


lesser = lambda x, y: x if x < y else y
greater = lambda x, y: x if x > y else y
countit = lambda acc, _: acc + 1


from operator import add, or_, and_

# Dict mapping
# Reduction : (binop, combiner, init)

# Reduction :: [a] -> b
# binop     :: b, a -> b
# combiner  :: b, b -> b
# init      :: b
binops = {reductions.sum: (add, add, 0),
          reductions.min: (lesser, lesser, 1e250),
          reductions.max: (greater, lesser, -1e250),
          reductions.count: (countit, add, 0),
          reductions.any: (or_, or_, False),
          reductions.all: (and_, and_, True)}


def child(expr):
    if len(expr._inputs) > 1:
        raise ValueError()
    return expr._inputs[0]


def reduce_by_funcs(t):
    """ Create grouping func and binary operator for a by-reduction/summary

    Turns a by operation like

        by(t.name, t.amount.sum())

    into a grouper like

    >>> def grouper(row):
    ...     return row[name_index]

    and a binary operator like

    >>> def binop(acc, row):
    ...     return binops[sum](acc, row[amount_index])

    It also handles this in the more complex ``summary`` case in which case
    several binary operators are juxtaposed together.

    See Also:
        compute_up(By, Sequence)
    """
    grouper = rrowfunc(t.grouper, t._child)
    if (isinstance(t.apply, Reduction) and
            type(t.apply) in binops):

        binop, combiner, initial = binops[type(t.apply)]
        applier = rrowfunc(t.apply._child, t._child)

        def binop2(acc, x):
            return binop(acc, applier(x))

        return grouper, binop2, combiner, initial

    elif (isinstance(t.apply, Summary) and
          builtins.all(type(val) in binops for val in t.apply.values)):

        binops2, combiners, inits = zip(
            *[binops[type(v)] for v in t.apply.values])
        appliers = [rrowfunc(v._child, t._child) for v in t.apply.values]

        def binop2(accs, x):
            return tuple(binop(acc, applier(x)) for binop, acc, applier in
                         zip(binops2, accs, appliers))

        def combiner(a, b):
            return tuple(c(x, y) for c, x, y in zip(combiners, a, b))

        return grouper, binop2, combiner, tuple(inits)


@dispatch(By, Sequence)
def compute_up(t, seq, **kwargs):
    apply = optimize(t.apply, seq)
    grouper = optimize(t.grouper, seq)
    t = By(grouper, apply)
    if ((isinstance(t.apply, Reduction) and type(t.apply) in binops) or
        (isinstance(t.apply, Summary) and builtins.all(type(val) in binops
                                                       for val in t.apply.values))):
        grouper, binop, combiner, initial = reduce_by_funcs(t)
        d = reduceby(grouper, binop, seq, initial)
    else:
        grouper = rrowfunc(t.grouper, t._child)
        groups = groupby(grouper, seq)
        d = dict((k, compute(t.apply, {t._child: v}, return_type='native'))
                 for k, v in groups.items())

    if isscalar(t.grouper.dshape.measure):
        keyfunc = lambda x: (x,)
    else:
        keyfunc = identity
    if isscalar(t.apply.dshape.measure):
        valfunc = lambda x: (x,)
    else:
        valfunc = identity
    return tuple(keyfunc(k) + valfunc(v) for k, v in d.items())


def pair_assemble(t, on_left=None, on_right=None):
    """ Combine a pair of records into a single record

    This is mindful to shared columns as well as missing records

    Parameters
    ----------
    t : Join
        The join to combine.
    on_left : list of column indicies, optional
        The column indicies of the left columns.
    on_right : list of column indicies, optional
        The column indicies of the right columns.

    Returns
    -------
    assemble : callable
        A function that assembles the data from a row of the
        left and right data sources.
    """
    try:
        from cytoolz import get  # not curried version
    except:
        from toolz import get
    on_left = (
        on_left
        if on_left is not None else
        [t.lhs.fields.index(col) for col in listpack(t.on_left)]
    )
    on_right = (
        on_right
        if on_right is not None else
        [t.rhs.fields.index(col) for col in listpack(t.on_right)]
    )

    left_self_columns = [t.lhs.fields.index(c) for c in t.lhs.fields
                         if c not in listpack(t.on_left)]
    right_self_columns = [t.rhs.fields.index(c) for c in t.rhs.fields
                          if c not in listpack(t.on_right)]

    def assemble(pair):
        a, b = pair
        if a is not None:
            joined = get(on_left, a)
        else:
            joined = get(on_right, b)

        if a is not None:
            left_entries = get(left_self_columns, a)
        else:
            left_entries = (None,) * (len(t.lhs.fields) - len(on_left))

        if b is not None:
            right_entries = get(right_self_columns, b)
        else:
            right_entries = (None,) * (len(t.rhs.fields) - len(on_right))

        return joined + left_entries + right_entries

    return assemble


@dispatch(Join, Sequence, Sequence)
def compute_up(t, lhs, rhs, **kwargs):
    """ Join Operation for Python Streaming Backend

    Note that a pure streaming Join is challenging/impossible because any row
    in one seq might connect to any row in the other, requiring simultaneous
    complete access.

    As a result this approach compromises and fully realizes the LEFT sequence
    while allowing the RIGHT sequence to stream.  As a result

    Always put your bigger collection on the RIGHT side of the Join.
    """
    if lhs == rhs:
        lhs, rhs = itertools.tee(lhs, 2)

    on_left = [t.lhs.fields.index(col) for col in listpack(t.on_left)]
    on_right = [t.rhs.fields.index(col) for col in listpack(t.on_right)]

    left_default = (None if t.how in ('right', 'outer')
                    else toolz.itertoolz.no_default)
    right_default = (None if t.how in ('left', 'outer')
                     else toolz.itertoolz.no_default)

    pairs = toolz.join(on_left, lhs,
                       on_right, rhs,
                       left_default=left_default,
                       right_default=right_default)

    assemble = pair_assemble(t, on_left, on_right)

    return map(assemble, pairs)


@dispatch(Sort, Sequence)
def compute_up(t, seq, **kwargs):
    if isscalar(t._child.dshape.measure) and t.key == t._child._name:
        key = identity
    elif isinstance(t.key, (str, unicode, tuple, list)):
        key = rowfunc(t._child[t.key])
    else:
        key = rrowfunc(optimize(t.key, seq), t._child)
    return sorted(seq,
                  key=key,
                  reverse=not t.ascending)


@dispatch(Head, Sequence)
def compute_up(t, seq, **kwargs):
    if t.n < 100:
        return tuple(take(t.n, seq))
    else:
        return take(t.n, seq)


@dispatch(Sample, Sequence)
def compute_up(t, seq, **kwargs):
    nsamp = t.n if t.n is not None else int(t.frac * len(seq))
    return random.sample(seq, min(nsamp, len(seq)))


@dispatch((Label, ReLabel), Sequence)
def compute_up(t, seq, **kwargs):
    return seq


@dispatch(Apply, Sequence)
def compute_up(t, seq, **kwargs):
    return t.func(seq)


@dispatch(Summary, Sequence)
def compute_up(expr, data, **kwargs):
    if expr._child.ndim != 1:
        raise NotImplementedError('Only 1D reductions currently supported')
    if isinstance(data, Iterator):
        datas = itertools.tee(data, len(expr.values))
        result = tuple(
            compute(val, {expr._child: data}, return_type='native')
            for val, data in zip(expr.values, datas)
        )
    else:
        result = tuple(
            compute(val, {expr._child: data}, return_type='native')
            for val in expr.values
        )

    if expr.keepdims:
        return (result,)
    else:
        return result


@dispatch(Like, Sequence)
def compute_up(expr, seq, **kwargs):
    def func(x, pattern=expr.pattern):
        return fnmatch.fnmatch(x, pattern)
    return map(func, seq)


@dispatch(Slice, Sequence)
def compute_up(expr, seq, **kwargs):
    index = expr.index
    if isinstance(index, tuple) and len(index) == 1:
        index = index[0]
    if isinstance(index, _inttypes):
        try:
            return seq[index]
        except:
            if index >= 0:
                return nth(index, seq)
            else:
                return tail(-index, seq)[0]
    if isinstance(index, slice):
        if (index.start and index.start < 0 and
                index.stop is None and
                index.step in (1, None)):
            return tail(-index.start, seq)
        else:
            return itertools.islice(seq, index.start, index.stop, index.step)
    raise NotImplementedError("Only 1d slices supported")


@dispatch(Coerce, (np.float32, np.float64, np.int64, np.int32, base))
def compute_up(expr, ob, **kwargs):
    tp = expr.to
    shape = tp.shape
    if shape:
        raise TypeError(
            'cannot convert scalar object %r to array or matrix shape %r' % (
                ob,
                shape,
            ),
        )

    measure = tp.measure
    if isinstance(measure, Option):
        if pd.isnull(ob):
            return None
        measure = measure.ty
    dtype = to_numpy_dtype(measure)
    return dtype.type(ob)



import pytest
from blaze.expr import (symbol, transform, by, count, summary, var, std, mean,
                        sqrt, sum)
from blaze.expr.split import split, path_split
from datashape import dshape
from datashape.predicates import isscalar, isrecord, iscollection

t = symbol('t', 'var * {name: string, amount: int32, id: int32}')
a = symbol('a', '1000 * 2000 * {x: float32, y: float32}')


def test_path_split():
    expr = t.amount.sum() + 1
    assert path_split(t, expr).isidentical(t.amount.sum())

    expr = t.amount.distinct().sort()
    assert path_split(t, expr).isidentical(t.amount.distinct())

    t2 = transform(t, id=t.id * 2)
    expr = by(t2.id, amount=t2.amount.sum()).amount + 1
    assert path_split(t, expr).isidentical(by(t2.id, amount=t2.amount.sum()))

    expr = count(t.amount.distinct())
    assert path_split(t, expr).isidentical(t.amount.distinct())

    expr = summary(total=t.amount.sum())
    assert path_split(t, expr).isidentical(expr)


def test_sum():
    (chunk, chunk_expr), (agg, agg_expr) = split(t, t.amount.sum())

    assert chunk.schema == t.schema
    assert chunk_expr.isidentical(chunk.amount.sum(keepdims=True))

    assert isscalar(agg.dshape.measure)
    assert agg_expr.isidentical(sum(agg))


def test_mean():
    (chunk, chunk_expr), (agg, agg_expr) = split(t, t.amount.mean())

    assert chunk.schema == t.schema
    assert chunk_expr.isidentical(summary(total=chunk.amount.sum(),
                                          count=chunk.amount.count(),
                                          keepdims=True))
    assert isrecord(agg.dshape.measure)
    assert agg_expr.isidentical(agg.total.sum() / agg['count'].sum())


def test_var():
    (chunk, chunk_expr), (agg, agg_expr) = split(t, t.amount.var())

    assert chunk.schema == t.schema
    assert chunk_expr.isidentical(summary(x=chunk.amount.sum(),
                                          x2=(chunk.amount ** 2).sum(),
                                          n=chunk.amount.count(),
                                          keepdims=True))

    assert isrecord(agg.dshape.measure)
    assert agg_expr.isidentical(agg.x2.sum() / (agg.n.sum()) -
                                (agg.x.sum() / (agg.n.sum())) ** 2)


def test_std():
    (chunk, chunk_expr), (agg, agg_expr) = split(t, t.amount.std())

    assert chunk.schema == t.schema
    assert chunk_expr.isidentical(summary(x=chunk.amount.sum(),
                                          x2=(chunk.amount ** 2).sum(),
                                          n=chunk.amount.count(),
                                          keepdims=True))

    assert isrecord(agg.dshape.measure)
    assert agg_expr.isidentical(sqrt(agg.x2.sum() / agg.n.sum() -
                                     (agg.x.sum() / (agg.n.sum())) ** 2))


def test_sum_with_axis_argument():
    chunk = symbol('chunk', '100 * 100 * {x: float32, y: float32}')
    (chunk, chunk_expr), (agg, agg_expr) = split(
        a, a.x.sum(axis=0), chunk=chunk)

    assert chunk.schema == a.schema
    assert agg_expr.dshape == a.x.sum(axis=0).dshape

    assert chunk_expr.isidentical(chunk.x.sum(axis=0, keepdims=True))
    assert agg_expr.isidentical(agg.sum(axis=0))


def test_sum_with_keepdims():
    (chunk, chunk_expr), (agg, agg_expr) = split(
        t, t.amount.sum(keepdims=True))

    assert chunk.schema == t.schema
    assert chunk_expr.isidentical(chunk.amount.sum(keepdims=True))

    assert isscalar(agg.dshape.measure)
    assert agg_expr.isidentical(sum(agg, keepdims=True))


def test_split_reasons_correctly_about_uneven_aggregate_shape():
    x = symbol('chunk', '10 * 10 * int')
    chunk = symbol('chunk', '3 * 3 * int')
    (chunk, chunk_expr), (agg, agg_expr) = split(x, x.sum(axis=0),
                                                 chunk=chunk)
    assert agg.shape == (4, 10)

    x = symbol('leaf', '1643 * 60 * int')
    chunk = symbol('chunk', '40 * 60 * int')
    (chunk, chunk_expr), (agg, agg_expr) = split(x, x.sum(),
                                                 chunk=chunk)
    assert agg.shape == (42, 1)


def test_split_reasons_correctly_about_aggregate_shape():
    chunk = symbol('chunk', '100 * 100 * {x: float32, y: float32}')
    (chunk, chunk_expr), (agg, agg_expr) = split(a, a.x.sum(), chunk=chunk)

    assert agg.shape == (10, 20)

    chunk = symbol('chunk', '100 * 100 * {x: float32, y: float32}')
    (chunk, chunk_expr), (agg, agg_expr) = split(
        a, a.x.sum(axis=0), chunk=chunk)

    assert agg.shape == (10, 2000)


def test_distinct():
    (chunk, chunk_expr), (agg, agg_expr) = split(t, count(t.amount.distinct()))

    assert chunk.schema == t.schema
    assert chunk_expr.isidentical(chunk.amount.distinct())

    assert isscalar(agg.dshape.measure)
    assert agg_expr.isidentical(count(agg.distinct()))


def test_summary():
    (chunk, chunk_expr), (agg, agg_expr) = split(t, summary(a=t.amount.count(),
                                                            b=t.id.sum() + 1))

    assert chunk.schema == t.schema
    assert chunk_expr.isidentical(summary(a=chunk.amount.count(),
                                          b=chunk.id.sum(),
                                          keepdims=True))

    # assert not agg.schema == dshape('{a: int32, b: int32}')
    assert agg_expr.isidentical(summary(a=agg.a.sum(),
                                        b=agg.b.sum() + 1))

    (chunk, chunk_expr), (agg, agg_expr) = split(t,
                                                 summary(total=t.amount.sum()))

    assert chunk_expr.isidentical(summary(total=chunk.amount.sum(),
                                          keepdims=True))
    assert agg_expr.isidentical(summary(total=agg.total.sum()))


def test_summary_with_mean():
    (chunk, chunk_expr), (agg, agg_expr) = split(t, summary(a=t.amount.count(),
                                                            b=t.id.mean() + 1))

    assert chunk.schema == t.schema
    assert chunk_expr.isidentical(summary(a=chunk.amount.count(),
                                          b_total=chunk.id.sum(),
                                          b_count=chunk.id.count(),
                                          keepdims=True))

    # assert not agg.schema == dshape('{a: int32, b: int32}')
    expected = summary(
        a=agg.a.sum(),
        b=(agg.b_total.sum() / agg.b_count.sum()) + 1,
    )
    assert agg_expr.isidentical(expected)


def test_complex_summaries():
    t = symbol('t', '100 * {a: int, b: int}')
    (chunk, chunk_expr), (agg, agg_expr) = split(t, summary(q=t.a.mean(),
                                                            w=t.a.std(),
                                                            e=t.a.sum()))

    assert chunk_expr.isidentical(summary(e=chunk.a.sum(),
                                          q_count=chunk.a.count(),
                                          q_total=chunk.a.sum(),
                                          w_n=chunk.a.count(),
                                          w_x=chunk.a.sum(),
                                          w_x2=(chunk.a ** 2).sum(),
                                          keepdims=True))

    expected = summary(
        e=agg.e.sum(),
        q=agg.q_total.sum() / agg.q_count.sum(),
        w=sqrt((agg.w_x2.sum() / agg.w_n.sum())
               - (agg.w_x.sum() / agg.w_n.sum()) ** 2),
    )
    assert agg_expr.isidentical(expected)


def test_by_sum():
    (chunk, chunk_expr), (agg, agg_expr) = \
        split(t, by(t.name, total=t.amount.sum()))

    assert chunk.schema == t.schema
    assert chunk_expr.isidentical(by(chunk.name, total=chunk.amount.sum()))

    assert not isscalar(agg.dshape.measure)
    assert agg_expr.isidentical(by(agg.name, total=agg.total.sum()))


def test_by_count():
    (chunk, chunk_expr), (agg, agg_expr) = \
        split(t, by(t.name, total=t.amount.count()))

    assert chunk_expr.isidentical(by(chunk.name, total=chunk.amount.count()))

    assert agg_expr.isidentical(by(agg.name, total=agg.total.sum()))


def test_by_mean():
    (chunk, chunk_expr), (agg, agg_expr) = \
        split(t, by(t.name, avg=t.amount.mean()))

    assert chunk_expr.isidentical(by(chunk.name,
                                     avg_total=chunk.amount.sum(),
                                     avg_count=chunk.amount.count()))

    assert agg_expr.isidentical(by(agg.name,
                                   avg=agg.avg_total.sum() /
                                   agg.avg_count.sum()))


def test_embarassing_rowwise():
    (chunk, chunk_expr), (agg, agg_expr) = split(t, t.amount + 1)

    assert chunk_expr.isidentical(chunk.amount + 1)
    assert agg_expr.isidentical(agg)


def test_embarassing_selection():
    (chunk, chunk_expr), (agg, agg_expr) = split(t, t[t.amount > 0])

    assert chunk_expr.isidentical(chunk[chunk.amount > 0])
    assert agg_expr.isidentical(agg)


def test_embarassing_like():
    (chunk, chunk_expr), (agg, agg_expr) = split(t, t[t.name.like('Alice*')])

    assert chunk_expr.isidentical(chunk[chunk.name.like('Alice*')])
    assert agg_expr.isidentical(agg)


x = symbol('x', '24 * 16 * int32')


def test_nd_chunk():
    c = symbol('c', '4 * 4 * int32')

    (chunk, chunk_expr), (agg, agg_expr) = split(x, x.sum(), chunk=c)

    assert chunk.shape == (4, 4)
    assert chunk_expr.isidentical(chunk.sum(keepdims=True))

    assert agg.shape == (6, 4)
    assert agg_expr.isidentical(agg.sum())


def test_nd_chunk_axis_args():
    c = symbol('c', '4 * 4 * int32')

    (chunk, chunk_expr), (agg, agg_expr) = split(x, x.sum(axis=0), chunk=c)

    assert chunk.shape == (4, 4)
    assert chunk_expr.shape == (1, 4)
    assert chunk_expr.isidentical(chunk.sum(keepdims=True, axis=0))

    assert agg.shape == (6, 16)
    assert agg_expr.isidentical(agg.sum(axis=0))

    for func in [var, std, mean]:
        (chunk, chunk_expr), (agg, agg_expr) = split(
            x, func(x, axis=0), chunk=c)

        assert chunk.shape == (4, 4)
        assert chunk_expr.shape == (1, 4)
        assert agg.shape == (6, 16)


def test_agg_shape_in_tabular_case_with_explicit_chunk():
    t = symbol('t', '1000 * {name: string, amount: int, id: int}')
    c = symbol('chunk', 100 * t.schema)

    expr = by(t.name, total=t.amount.sum())
    (chunk, chunk_expr), (agg, agg_expr) = split(t, expr, chunk=c)

    assert agg.dshape == dshape('var * {name: string, total: int64}')


def test_reductions():
    (chunk, chunk_expr), (agg, agg_expr) = split(t, t.amount.nunique())

    assert chunk.schema == t.schema
    assert chunk_expr.isidentical(chunk.amount.distinct())

    assert isscalar(agg.dshape.measure)
    assert agg_expr.isidentical(agg.distinct().count())

    (chunk, chunk_expr), (agg, agg_expr) = \
        split(t, t.amount.nunique(keepdims=True))

    assert chunk.schema == t.schema
    assert chunk_expr.isidentical(chunk.amount.distinct())

    assert isscalar(agg.dshape.measure)
    assert agg_expr.isidentical(agg.distinct().count(keepdims=True))


def test_by_with_single_field_child():
    x = symbol('x', 'var * int')
    (chunk, chunk_expr), (agg, agg_expr) = split(x, by(x, total=x.sum()))

    assert chunk_expr.isidentical(by(chunk, total=chunk.sum()))

    assert agg_expr.isidentical(by(
        agg[agg.fields[0]],
        total=agg.total.sum()
    ).relabel({agg.fields[0]: 'x'}))


def test_keepdims_equals_true_doesnt_mess_up_agg_shape():
    x = symbol('x', '10 * int')
    (chunk, chunk_expr), (agg, agg_expr) = split(x, x.sum(), keepdims=False)

    assert iscollection(agg.dshape)


def test_splittable_apply():
    def f(x):
        pass

    (chunk, chunk_expr), (agg, agg_expr) = \
        split(t, t.amount.apply(f, 'var * int', splittable=True))
    assert chunk_expr.isidentical(
        chunk.amount.apply(f, 'var * int', splittable=True),
    )

    assert agg_expr.isidentical(agg)


@pytest.mark.xfail(raises=AssertionError,
                   reason="Can't split on expressions requiring branching")
def test_elemwise_with_multiple_paths():
    s = symbol('s', 'var * {x: int, y: int, z: int}')
    expr = s.x.sum() / s.y.sum()

    (chunk, chunk_expr), (agg, agg_expr) = split(s, expr)
    assert chunk_expr.isidentical(summary(x=chunk.x.sum(), y=chunk.y.sum()))
    assert agg_expr.isidentical(agg.x / agg.y)



from __future__ import absolute_import, division, print_function

import re

from .core import common_subexpression
from .expressions import Expr
from .reductions import Reduction, Summary, summary
from ..dispatch import dispatch
from .expressions import dshape_method_list

from datashape import dshape, Record, Map, Unit, var

__all__ = ['by', 'By', 'count_values']


def _names_and_types(expr):
    schema = expr.dshape.measure
    schema = getattr(schema, 'ty', schema)
    if isinstance(schema, Record):
        return schema.names, schema.types
    if isinstance(schema, Unit):
        return [expr._name], [expr.dshape.measure]
    if isinstance(schema, Map):
        return [expr._name], [expr.dshape.measure.key]
    raise ValueError("Unable to determine name and type of %s" % expr)


class By(Expr):
    """ Split-Apply-Combine Operator

    Examples
    --------

    >>> from blaze import symbol
    >>> t = symbol('t', 'var * {name: string, amount: int, id: int}')
    >>> e = by(t['name'], total=t['amount'].sum())

    >>> data = [['Alice', 100, 1],
    ...         ['Bob', 200, 2],
    ...         ['Alice', 50, 3]]

    >>> from blaze.compute.python import compute
    >>> sorted(compute(e, data))
    [('Alice', 150), ('Bob', 200)]
    """

    _arguments = 'grouper', 'apply'

    @property
    def _child(self):
        return common_subexpression(self.grouper, self.apply)

    def _schema(self):
        grouper_names, grouper_types = _names_and_types(self.grouper)
        apply_names, apply_types = _names_and_types(self.apply)

        names = grouper_names + apply_names
        types = grouper_types + apply_types

        return dshape(Record(list(zip(names, types))))

    def _dshape(self):
        # TODO: think if this should be generalized
        return var * self.schema

    def __str__(self):
        return '%s(%s, %s)' % (type(self).__name__.lower(),
                               self.grouper,
                               re.sub(r'^summary\((.*)\)$', r'\1',
                                      str(self.apply)))


@dispatch(Expr, Reduction)
def by(grouper, s):
    raise ValueError("This syntax has been removed.\n"
                     "Please name reductions with keyword arguments.\n"
                     "Before:   by(t.name, t.amount.sum())\n"
                     "After:    by(t.name, total=t.amount.sum())")


@dispatch(Expr, Summary)
def by(grouper, s):
    return By(grouper, s)


@dispatch(Expr)
def by(grouper, **kwargs):
    return By(grouper, summary(**kwargs))


def count_values(expr, sort=True):
    """
    Count occurrences of elements in this column

    Sort by counts by default
    Add ``sort=False`` keyword to avoid this behavior.
    """
    result = by(expr, count=expr.count())
    if sort:
        result = result.sort('count', ascending=False)
    return result


dshape_method_list.extend([
    (lambda ds: len(ds.shape) == 1, set([count_values])),
    ])



from __future__ import absolute_import, division, print_function

import sys
import logging
from logging import Formatter
from functools import wraps
import traceback
import collections
from datetime import datetime
import errno
import functools
from hashlib import md5
import os
import socket
from time import time
from warnings import warn
import importlib

from datashape import discover, pprint
import flask
from flask import Blueprint, Flask, Response
from flask_cors import cross_origin
from werkzeug.http import parse_options_header
from toolz import valmap, compose

import blaze
from blaze import compute, resource
from blaze.compatibility import ExitStack, u8
from blaze.compute import compute_up
from .serialization import json, all_formats
from ..interactive import _Data
from ..expr import Expr, symbol, utils as expr_utils, Symbol


__all__ = 'Server', 'to_tree', 'from_tree', 'expr_md5'

# http://www.speedguide.net/port.php?port=6363
# http://en.wikipedia.org/wiki/List_of_TCP_and_UDP_port_numbers
DEFAULT_PORT = 6363


class RC(object):
    """
    Simple namespace for HTTP status codes.
    https://en.wikipedia.org/wiki/List_of_HTTP_status_codes
    """

    OK = 200
    CREATED = 201

    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    NOT_FOUND = 404
    FORBIDDEN = 403
    CONFLICT = 409
    UNPROCESSABLE_ENTITY = 422
    UNSUPPORTED_MEDIA_TYPE = 415

    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501


api = Blueprint('api', __name__)
pickle_extension_api = Blueprint('pickle_extension_api', __name__)


_no_default = object()  # sentinel


def _logging(func):
    @wraps(func)
    def _logger(*args, **kwargs):
        logger = flask.current_app.logger
        try:
            logger.debug("Calling %s" % func.__name__)
            ret = func(*args, **kwargs)
        finally:
            logger.debug("Leaving %s" % func.__name__)
        return ret
    return _logger


def _get_option(option, options, default=_no_default):
    try:
        return options[option]
    except KeyError:
        if default is not _no_default:
            return default

        # Provides a more informative error message.
        msg = 'The blaze api must be registered with {option}'
        raise TypeError(msg.format(option=option))


def ensure_dir(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

# Default for logging exception tracebacks is to simply use the standard
# `traceback.format_tb`
_default_log_exception_formatter = compose(''.join, traceback.format_tb)


def _register_api(app, options, first_registration=False):
    """
    Register the data with the blueprint.
    """
    _get_data.cache[app] = _get_option('data', options)
    _get_format.cache[app] = {f.name: f for f in _get_option('formats', options)}
    _get_auth.cache[app] = (_get_option('authorization', options, None) or
                            (lambda a: True))
    allow_profiler = _get_option('allow_profiler', options, False)
    profiler_output = _get_option('profiler_output', options, None)
    profile_by_default = _get_option('profile_by_default', options, False)
    if not allow_profiler and (profiler_output or profile_by_default):
        msg = "cannot set %s%s%s when 'allow_profiler' is False"
        raise ValueError(msg % ('profiler_output' if profiler_output else '',
                                ' or ' if profiler_output and profile_by_default else '',
                                'profile_by_default' if profile_by_default else ''))
    if allow_profiler:
        if profiler_output is None:
            profiler_output = 'profiler_output'
        if profiler_output != ':response':
            ensure_dir(profiler_output)

    _get_profiler_info.cache[app] = (allow_profiler,
                                     profiler_output,
                                     profile_by_default)

    # Allowing users to dynamically add datasets to the Blaze server can be
    # dangerous, so we only expose the method if specifically requested
    allow_add = _get_option('allow_add', options, False)
    if allow_add:
        app.add_url_rule('/add', 'addserver', addserver,
                         methods=['POST', 'HEAD', 'OPTIONS'])

    # Call the original register function.
    Blueprint.register(api, app, options, first_registration)

api.register = _register_api


def per_app_accesor(name):
    def _get():
        return _get.cache[flask.current_app]
    _get.cache = {}
    _get.__name__ = '_get' + name
    return _get


def _get_format(name):
    return _get_format.cache[flask.current_app][name]
_get_format.cache = {}

_get_data = per_app_accesor('data')
_get_auth = per_app_accesor('auth')
_get_profiler_info = per_app_accesor('profiler_info')


def expr_md5(expr):
    """Returns the md5 hash of the str of the expression.

    Parameters
    ----------
    expr : Expr
        The expression to hash.

    Returns
    -------
    hexdigest : str
        The hexdigest of the md5 of the str of ``expr``.
    """
    exprstr = str(expr)
    if not isinstance(exprstr, bytes):
        exprstr = exprstr.encode('utf-8')
    return md5(exprstr).hexdigest()


def _prof_path(profiler_output, expr):
    """Get the path to write the data for a profile run of ``expr``.

    Parameters
    ----------
    profiler_output : str
        The director to write into.
    expr : Expr
        The expression that was run.

    Returns
    -------
    prof_path : str
        The filepath to write the new profiler data.

    Notes
    -----
    This function ensures that the dirname of the returned path exists.
    """
    dir_ = os.path.join(profiler_output,
                        expr_md5(expr))  # Use md5 so the client knows where to look.
    ensure_dir(dir_)
    return os.path.join(dir_,
                        str(int(datetime.utcnow().timestamp())))


def authorization(f):
    @functools.wraps(f)
    def authorized(*args, **kwargs):
        if not _get_auth()(flask.request.authorization):
            return Response('bad auth token',
                            RC.UNAUTHORIZED,
                            {'WWW-Authenticate': 'Basic realm="Login Required"'})
        return f(*args, **kwargs)
    return authorized


def check_request(f):
    @functools.wraps(f)
    def check():
        raw_content_type = flask.request.headers['content-type']
        content_type, options = parse_options_header(raw_content_type)

        if content_type not in accepted_mimetypes:
            return ('Unsupported serialization format %s' % content_type,
                    RC.UNSUPPORTED_MEDIA_TYPE)

        try:
            serial = _get_format(accepted_mimetypes[content_type])
        except KeyError:
            return ("Unsupported serialization format '%s'" % content_type,
                    RC.UNSUPPORTED_MEDIA_TYPE)

        try:
            payload = serial.loads(flask.request.data)
        except ValueError:
            return ("Bad data.  Got %s " % flask.request.data, RC.BAD_REQUEST)

        return f(payload, serial)
    return check


class FlaskWithExceptionFormatting(Flask):
    """ Add a `log_exception_formatter` instance attribute to the Flask
    application object, to allow it to store a handler function.
    """
    log_exception_formatter = None

    def __init__(self, *args, **kwargs):
        self.log_exception_formatter = kwargs.pop('log_exception_formatter',
                                                  _default_log_exception_formatter)
        super(FlaskWithExceptionFormatting, self).__init__(*args, **kwargs)


class Server(object):

    """ Blaze Data Server

    Host local data through a web API

    Parameters
    ----------
    data : dict, optional
        A dictionary mapping dataset name to any data format that blaze
        understands.
    formats : iterable, optional
        An iterable of supported serialization formats. By default, the
        server will support JSON.
        A serialization format is an object that supports:
        name, loads, and dumps.
    authorization : callable, optional
        A callable to be used to check the auth header from the client.
        This callable should accept a single argument that will either be
        None indicating that no header was passed, or an object
        containing a username and password attribute. By default, all requests
        are allowed.
    allow_profiler : bool, optional
        Allow payloads to specify `"profile": true` which will run the
        computation under cProfile.
    profiler_output : str, optional
        The directory to write pstats files after profile runs.
        The files will be written in a structure like:

          {profiler_output}/{hash(expr)}/{timestamp}

        This defaults to a relative path of `profiler_output`.
        This requires `allow_profiler=True`.

        If this is the string ':response' then writing to the local filesystem
        is disabled. Only requests that specify `profiler_output=':response'`
        will be served. All others will return a 403 (Forbidden).
    profile_by_default : bool, optional
        Run the profiler on any computation that does not explicitly set
        "profile": false.
        This requires `allow_profiler=True`.
    allow_add : bool, optional
        Expose an `/add` endpoint to allow datasets to be dynamically added to
        the server. Since this increases the risk of security holes, it defaults
        to `False`.
    logfile : str or file-like object, optional
        A filename or open file-like stream to which to send log output. Defaults
        to `sys.stdout`.
    loglevel : str, optional
        A string logging level (e.g. 'WARNING', 'INFO') to set how verbose log
        output should be.
    log_exception_formatter : callable, optional
        A callable to be used to format an exception traceback for logging. It
        should take a traceback argument, and return the string to be logged.
        This defaults to the standard library `traceback.format_tb`


    Examples
    --------
    >>> from pandas import DataFrame
    >>> df = DataFrame([[1, 'Alice',   100],
    ...                 [2, 'Bob',    -200],
    ...                 [3, 'Alice',   300],
    ...                 [4, 'Dennis',  400],
    ...                 [5,  'Bob',   -500]],
    ...                columns=['id', 'name', 'amount'])

    >>> server = Server({'accounts': df})
    >>> server.run() # doctest: +SKIP
    """
    def __init__(self,
                 data=None,
                 formats=None,
                 authorization=None,
                 allow_profiler=False,
                 profiler_output=None,
                 profile_by_default=False,
                 allow_add=False,
                 logfile=sys.stdout,
                 loglevel='WARNING',
                 log_exception_formatter=_default_log_exception_formatter):
        if isinstance(data, collections.Mapping):
            data = valmap(lambda v: v.data if isinstance(v, _Data) else v,
                          data)
        elif isinstance(data, _Data):
            data = data._resources()
        app = self.app = FlaskWithExceptionFormatting('blaze.server.server',
                                                      log_exception_formatter=log_exception_formatter)
        if data is None:
            data = {}
        app.register_blueprint(api,
                               data=data,
                               formats=formats if formats is not None else (json,),
                               authorization=authorization,
                               allow_profiler=allow_profiler,
                               profiler_output=profiler_output,
                               profile_by_default=profile_by_default,
                               allow_add=allow_add)
        self.data = data
        if logfile:
            if isinstance(logfile, (str, bytes)):
                handler = logging.FileHandler(logfile)
            else:
                handler = logging.StreamHandler(logfile)
            handler.setFormatter(Formatter('[%(asctime)s %(levelname)s] %(message)s '
                                           '[in %(pathname)s:%(lineno)d]'))
            handler.setLevel(getattr(logging, loglevel))
            app.logger.addHandler(handler)

    def run(self, port=DEFAULT_PORT, retry=False, **kwargs):
        """Run the server.

        Parameters
        ----------
        port : int, optional
            The port to bind to.
        retry : bool, optional
            If the port is busy, should we retry with the next available port?
        **kwargs
            Forwarded to the underlying flask app's ``run`` method.

        Notes
        -----
        This function blocks forever when successful.
        """
        self.port = port
        try:
            # Blocks until the server is shut down.
            self.app.logger.debug('Starting server...')
            self.app.run(port=port, **kwargs)
            self.app.logger.debug('Stopping server...')
        except socket.error:
            if not retry:
                raise

            warn("Oops, couldn't connect on port %d.  Is it busy?" % port)
            # Attempt to start the server on a new port.
            self.run(port=port + 1, retry=retry, **kwargs)


@api.route('/datashape', methods=['GET'])
@cross_origin(origins='*', methods=['GET'])
@authorization
@_logging
def shape():
    return pprint(discover(_get_data()), width=0)


def to_tree(expr, names=None):
    """ Represent Blaze expression with core data structures

    Transform a Blaze expression into a form using only strings, dicts, lists
    and base types (int, float, datetime, ....)  This form can be useful for
    serialization.

    Parameters
    ----------
    expr : Expr
        A Blaze expression

    Examples
    --------

    >>> t = symbol('t', 'var * {x: int32, y: int32}')
    >>> to_tree(t) # doctest: +SKIP
    {'op': 'Symbol',
     'args': ['t', 'var * { x : int32, y : int32 }', False]}


    >>> to_tree(t.x.sum()) # doctest: +SKIP
    {'op': 'sum',
     'args': [{'op': 'Column',
               'args': [{'op': 'Symbol'
                         'args': ['t',
                                  'var * { x : int32, y : int32 }',
                                  False]}
                        'x']}]}

    Simplify expresion using explicit ``names`` dictionary.  In the example
    below we replace the ``Symbol`` node with the string ``'t'``.

    >>> tree = to_tree(t.x, names={t: 't'})
    >>> tree # doctest: +SKIP
    {'op': 'Column', 'args': ['t', 'x']}

    >>> from_tree(tree, namespace={'t': t})
    t.x

    See Also
    --------

    from_tree
    """
    if isinstance(expr, slice):
        # NOTE: This case must come first, since `slice` objects are not
        # hashable, so a dict lookup inside `names` will raise an execption.
        return {u'op': 'slice',
                u'args': [to_tree(arg, names=names) for arg in
                         [expr.start, expr.stop, expr.step]]}
    if names and expr in names:
        return names[expr]
    if isinstance(expr, tuple):
        return [to_tree(arg, names=names) for arg in expr]
    if isinstance(expr, expr_utils._slice):
        return to_tree(expr.as_slice(), names=names)
    elif isinstance(expr, _Data):
        return to_tree(symbol(u8(expr._name), expr.dshape), names)
    elif isinstance(expr, Expr):
        return {u'op': u8(type(expr).__name__),
                u'args': [to_tree(arg, names) for arg in expr._args]}
    elif isinstance(expr, str):
        return u8(expr)
    else:
        return expr


def expression_from_name(name):
    """

    >>> expression_from_name('By')
    <class 'blaze.expr.split_apply_combine.By'>

    >>> expression_from_name('And')
    <class 'blaze.expr.arithmetic.And'>
    """
    import blaze
    if hasattr(blaze, name):
        return getattr(blaze, name)
    if hasattr(blaze.expr, name):
        return getattr(blaze.expr, name)
    for signature, func in compute_up.funcs.items():
        try:
            if signature[0].__name__ == name:
                return signature[0]
        except TypeError:
            pass
    raise ValueError('%s not found in compute_up' % name)


def from_tree(expr, namespace=None):
    """ Convert core data structures to Blaze expression

    Core data structure representations created by ``to_tree`` are converted
    back into Blaze expressions.

    Parameters
    ----------
    expr : dict

    Examples
    --------

    >>> t = symbol('t', 'var * {x: int32, y: int32}')
    >>> tree = to_tree(t)
    >>> tree # doctest: +SKIP
    {'op': 'Symbol',
     'args': ['t', 'var * { x : int32, y : int32 }', False]}

    >>> from_tree(tree)
    <`t` symbol; dshape='var * {x: int32, y: int32}'>

    >>> tree = to_tree(t.x.sum())
    >>> tree # doctest: +SKIP
    {'op': 'sum',
     'args': [{'op': 'Field',
               'args': [{'op': 'Symbol'
                         'args': ['t',
                                  'var * {x : int32, y : int32}',
                                  False]}
                        'x']}]}

    >>> from_tree(tree)
    sum(t.x)

    Simplify expresion using explicit ``names`` dictionary.  In the example
    below we replace the ``Symbol`` node with the string ``'t'``.

    >>> tree = to_tree(t.x, names={t: 't'})
    >>> tree # doctest: +SKIP
    {'op': 'Field', 'args': ['t', 'x']}

    >>> from_tree(tree, namespace={'t': t})
    t.x

    See Also
    --------

    to_tree
    """
    if isinstance(expr, dict):
        op, args = expr[u'op'], expr[u'args']
        if 'slice' == op:
            return expr_utils._slice(*[from_tree(arg, namespace)
                                       for arg in args])
        if hasattr(blaze.expr, op):
            cls = getattr(blaze.expr, op)
        else:
            cls = expression_from_name(op)
        if cls is Symbol:
            cls = symbol
        children = [from_tree(arg, namespace) for arg in args]
        return cls(*children)
    elif isinstance(expr, (list, tuple)):
        return tuple(from_tree(arg, namespace) for arg in expr)
    if namespace and expr in namespace:
        return namespace[expr]
    else:
        return expr


accepted_mimetypes = {'application/vnd.blaze+{}'.format(x.name): x.name for x
                      in all_formats}


@api.route('/compute', methods=['POST', 'HEAD', 'OPTIONS'])
@cross_origin(origins='*', methods=['POST', 'HEAD', 'OPTIONS'])
@authorization
@check_request
@_logging
def compserver(payload, serial):
    expected_keys = {u'namespace',
                     u'odo_kwargs',
                     u'compute_kwargs',
                     u'expr',
                     u'profile',
                     u'profiler_output'}
    if not set(payload.keys()) < expected_keys:
        return ('unexpected keys in payload: %r' % sorted(set(payload.keys()) -
                                                          expected_keys),
                RC.BAD_REQUEST)

    app = flask.current_app
    (allow_profiler,
     default_profiler_output,
     profile_by_default) = _get_profiler_info()
    requested_profiler_output = payload.get(u'profiler_output',
                                            default_profiler_output)
    profile = payload.get(u'profile')
    profiling = (allow_profiler and
                 (profile or (profile_by_default and requested_profiler_output)))
    if profile and not allow_profiler:
        return ('profiling is disabled on this server', RC.FORBIDDEN)

    with ExitStack() as response_construction_context_stack:
        if profiling:
            from cProfile import Profile

            if (default_profiler_output == ':response' and
                    requested_profiler_output != ':response'):
                # writing to the local filesystem is disabled
                return ("local filepaths are disabled on this server, only"
                        " ':response' is allowed for the 'profiler_output' field",
                        RC.FORBIDDEN)

            profiler_output = requested_profiler_output
            profiler = Profile()
            profiler.enable()
            # ensure that we stop profiling in the case of an exception
            response_construction_context_stack.callback(profiler.disable)

        expr = '<failed to parse expr>'

        @response_construction_context_stack.callback
        def log_time(start=time()):
            app.logger.info('compute expr: %s\ntotal time (s): %.3f',
                            expr,
                            time() - start)

        ns = payload.get(u'namespace', {})
        compute_kwargs = payload.get(u'compute_kwargs') or {}
        odo_kwargs = payload.get(u'odo_kwargs') or {}
        dataset = _get_data()
        ns[':leaf'] = symbol('leaf', discover(dataset))

        expr = from_tree(payload[u'expr'], namespace=ns)

        if len(expr._leaves()) != 1:
            return ('too many leaves, expected 1 got %d' % len(expr._leaves()),
                    RC.BAD_REQUEST)

        leaf = expr._leaves()[0]

        formatter = getattr(flask.current_app, 'log_exception_formatter',
                            _default_log_exception_formatter)
        try:
            result = serial.materialize(compute(expr,
                                                {leaf: dataset},
                                                **compute_kwargs),
                                        expr.dshape,
                                        odo_kwargs)
        except NotImplementedError as e:
            # Note: `sys.exc_info()[2]` holds the current traceback, for
            # Python 2 / 3 compatibility. It's important not to store a local
            # reference to it.
            formatted_tb = formatter(sys.exc_info()[2])
            error_msg = "Computation not supported:\n%s\n%s" % (e, formatted_tb)
            app.logger.error(error_msg)
            return (error_msg, RC.NOT_IMPLEMENTED)
        except Exception as e:
            formatted_tb = formatter(sys.exc_info()[2])
            error_msg = "Computation failed with message:\n%s: %s\n%s" % (type(e).__name__, e, formatted_tb)
            app.logger.error(error_msg)
            return (error_msg, RC.INTERNAL_SERVER_ERROR)

        response = {u'datashape': pprint(expr.dshape, width=0),
                    u'data': serial.data_dumps(result),
                    u'names': expr.fields}

    if profiling:
        import marshal
        from pstats import Stats

        if profiler_output == ':response':
            from pandas.compat import BytesIO
            file = BytesIO()
        else:
            file = open(_prof_path(profiler_output, expr), 'wb')

        with file:
            # Use marshal to dump the stats data to the given file.
            # This is taken from cProfile which unfortunately does not have
            # an api that allows us to pass the file object directly, only
            # a file path.
            marshal.dump(Stats(profiler).stats, file)
            if profiler_output == ':response':
                response[u'profiler_output'] = {'__!bytes': file.getvalue()}

    return serial.dumps(response)


@cross_origin(origins='*', methods=['POST', 'HEAD', 'OPTIONS'])
@authorization
@check_request
@_logging
def addserver(payload, serial):
    """Add a data resource to the server.

    The request should contain serialized MutableMapping (dictionary) like
    object, and the server should already be hosting a MutableMapping resource.
    """

    data = _get_data.cache[flask.current_app]

    if not isinstance(data, collections.MutableMapping):
        data_not_mm_msg = ("Cannot update blaze server data since its current "
                           "data is a %s and not a mutable mapping (dictionary "
                           "like).")
        return (data_not_mm_msg % type(data), RC.UNPROCESSABLE_ENTITY)

    if not isinstance(payload, collections.Mapping):
        payload_not_mm_msg = ("Need a dictionary-like payload; instead was "
                              "given %s of type %s.")
        return (payload_not_mm_msg % (payload, type(payload)),
                RC.UNPROCESSABLE_ENTITY)

    if len(payload) > 1:
        error_msg = "Given more than one resource to add: %s"
        return (error_msg % list(payload.keys()),
                RC.UNPROCESSABLE_ENTITY)

    [(name, resource_info)] = payload.items()
    flask.current_app.logger.debug("Attempting to add dataset '%s'" % name)

    if name in data:
        msg = "Cannot add dataset named %s, already exists on server."
        return (msg % name, RC.CONFLICT)

    try:
        imports = []
        if isinstance(resource_info, dict):
            # Extract resource creation arguments
            source = resource_info['source']
            imports = resource_info.get('imports', [])
            args = resource_info.get('args', [])
            kwargs = resource_info.get('kwargs', {})
        else:
            # Just a URI
            source, args, kwargs = resource_info, [], {}
        # If we've been given libraries to import, we need to do so
        # before we can create the resource.
        for mod in imports:
            importlib.import_module(mod)
        # Make a new resource and try to discover it.
        new_resource = {name: resource(source, *args, **kwargs)}
        # Discovery is a minimal consistency check to determine if the new
        # resource is valid.
        ds = discover(new_resource)
        if name not in ds.dict:
            raise ValueError("%s not added." % name)
    except NotImplementedError as e:
        error_msg = "Addition not supported:\n%s: %s"
        return (error_msg % (type(e).__name__, e),
                RC.UNPROCESSABLE_ENTITY)
    except Exception as e:
        error_msg = "Addition failed with message:\n%s: %s"
        return (error_msg % (type(e).__name__, e),
                RC.UNPROCESSABLE_ENTITY)
    else:
        # Now that we've established that the new resource is discoverable--and
        # thus exists and is accessible--we add the resource to the server.
        data.update(new_resource)

    return ('OK', RC.CREATED)



"""
Dispatches functions like ``sum`` to builtins, numpy, or blaze depending on
input

>>> from blaze import sum, symbol

>>> sum([1, 2, 3])
6

>>> type(sum([1, 2, 3])).__name__
'int'

>>> type(sum(np.array([1, 2, 3], dtype=np.int64))).__name__
'int64'

>>> t = symbol('t', 'var * {x: int, y: int}')
>>> type(sum(t.x)).__name__
'sum'

"""
from __future__ import absolute_import, division, print_function

import math as pymath
from numbers import Number
import numpy as np
import pandas as pd

from multipledispatch import Dispatcher
from ..dispatch import namespace
from ..compatibility import builtins
from . import reductions
from . import math as blazemath
from .core import base
from .expressions import Expr, coalesce


"""
The following code creates reductions equivalent to the following:

@dispatch(Expr)
def sum(expr):
    return blaze.sum(expr)

@dispatch(np.ndarray)
def sum(x):
    return np.sum(x)

@dispatch(object)
def sum(o):
    return builtins.sum(o)


As well as mathematical functions like the following

@dispatch(Expr)
def sqrt(expr):
    return blaze.expr.math.sqrt(expr)

@dispatch(np.ndarray)
def sqrt(x):
    return np.sqrt(x)

@dispatch(object)
def sqrt(o):
    return math.sqrt(o)
"""

math_names = '''abs sqrt sin cos tan sinh cosh tanh acos acosh asin asinh atan atanh
exp log expm1 log10 log1p radians degrees ceil floor trunc isnan'''.split()

binary_math_names = "atan2 copysign fmod hypot ldexp greatest least coalesce".split()

reduction_names = '''any all sum min max mean var std'''.split()

__all__ = math_names + binary_math_names + reduction_names


types = {
    builtins: object,
    np: (np.ndarray, np.number),
    pymath: Number,
    blazemath: Expr,
    reductions: Expr,
}

binary_types = {
    builtins: [(object, object)],
    np: [((np.ndarray, np.number), (np.ndarray, np.number))],
    pymath: [(Number, Number)],
    blazemath: [(Expr, (Expr, base)), (base, Expr)]
}


def _coalesce_objects(lhs, rhs):
    # use pd.isnull for None, NaT, and others
    return lhs if not pd.isnull(lhs) else rhs


def _coalesce_arrays(lhs, rhs):
    np.where(pd.isnull(lhs), rhs, lhs)


fallback_binary_mappings = {
    'greatest': {
        builtins: max,
        np: np.maximum,
        pymath: max,
    },
    'least': {
        builtins: min,
        np: np.minimum,
        pymath: min,
    },
    'coalesce': {
        builtins: _coalesce_objects,
        np: _coalesce_arrays,
        pymath: _coalesce_objects,
        blazemath: coalesce,
    },
}


for funcname in math_names:  # sin, sqrt, ceil, ...
    d = Dispatcher(funcname)

    for module, typ in types.items():
        if hasattr(module, funcname):
            d.add((typ,), getattr(module, funcname))

    namespace[funcname] = d
    locals()[funcname] = d


for funcname in binary_math_names:  # hypot, atan2, fmod, ...
    d = Dispatcher(funcname)

    for module, pairs in binary_types.items():
        for pair in pairs:
            if hasattr(module, funcname):
                d.add(pair, getattr(module, funcname))
            elif funcname in fallback_binary_mappings:
                assert module in fallback_binary_mappings[funcname], module.__name__
                d.add(pair, fallback_binary_mappings[funcname][module])

    namespace[funcname] = d
    locals()[funcname] = d


for funcname in reduction_names:  # any, all, sum, max, ...
    d = Dispatcher(funcname)

    for module, typ in types.items():
        if hasattr(module, funcname):
            d.add((typ,), getattr(module, funcname))

    namespace[funcname] = d
    locals()[funcname] = d



from __future__ import print_function, division, absolute_import
import warnings
from functools import wraps


def deprecated(version, replacement=None):
    """Define a deprecation decorator.
    An optional `replacement` should refer to the new API to be used instead.

    Example:
      @deprecated('1.1')
      def old_func(): ...

      @deprecated('1.1', 'new_func')
      def old_func(): ..."""

    def decorator(func):
        def wrapper(*args, **kwargs):
            msg = "\"{}\" has been deprecated in version {} and will be removed in a future version."
            if replacement:
                msg += "\n Use \"{}\" instead."
            warnings.warn(msg.format(func.__name__, version, replacement),
                          category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wraps(func)(wrapper)

    return decorator



from __future__ import absolute_import, division, print_function

import pytest

import itertools

import numpy as np
import pandas as pd

from datetime import datetime, date

from blaze.compute.core import compute, compute_up
from blaze.expr import symbol, by, exp, summary, Broadcast, join, concat
from blaze.expr import greatest, least, coalesce
from blaze import sin
import blaze
from odo import into
from datashape import discover, to_numpy, dshape


x = np.array([(1, 'Alice', 100),
              (2, 'Bob', -200),
              (3, 'Charlie', 300),
              (4, 'Denis', 400),
              (5, 'Edith', -500)],
             dtype=[('id', 'i8'), ('name', 'S7'), ('amount', 'i8')])

t = symbol('t', discover(x))


def eq(a, b):
    c = a == b
    if isinstance(c, np.ndarray):
        return c.all()
    return c


def test_symbol():
    assert eq(compute(t, x), x)


def test_eq():
    assert eq(compute(t['amount'] == 100, x),
              x['amount'] == 100)


def test_selection():
    assert eq(compute(t[t['amount'] == 100], x), x[x['amount'] == 0])
    assert eq(compute(t[t['amount'] < 0], x), x[x['amount'] < 0])


def test_arithmetic():
    assert eq(compute(t['amount'] + t['id'], x),
              x['amount'] + x['id'])
    assert eq(compute(t['amount'] * t['id'], x),
              x['amount'] * x['id'])
    assert eq(compute(t['amount'] % t['id'], x),
              x['amount'] % x['id'])


def test_UnaryOp():
    assert eq(compute(exp(t['amount']), x),
              np.exp(x['amount']))

    assert eq(compute(abs(-t['amount']), x),
              abs(-x['amount']))


def test_Neg():
    assert eq(compute(-t['amount'], x),
              -x['amount'])


def test_invert_not():
    assert eq(compute(~(t.amount > 0), x),
              ~(x['amount'] > 0))


def test_Reductions():
    assert compute(t['amount'].mean(), x) == x['amount'].mean()
    assert compute(t['amount'].count(), x) == len(x['amount'])
    assert compute(t['amount'].sum(), x) == x['amount'].sum()
    assert compute(t['amount'].min(), x) == x['amount'].min()
    assert compute(t['amount'].max(), x) == x['amount'].max()
    assert compute(t['amount'].nunique(), x) == len(np.unique(x['amount']))
    assert compute(t['amount'].var(), x) == x['amount'].var()
    assert compute(t['amount'].std(), x) == x['amount'].std()
    assert compute(t['amount'].var(unbiased=True), x) == x['amount'].var(ddof=1)
    assert compute(t['amount'].std(unbiased=True), x) == x['amount'].std(ddof=1)
    assert compute((t['amount'] > 150).any(), x) == True
    assert compute((t['amount'] > 250).all(), x) == False
    assert compute(t['amount'][0], x) == x['amount'][0]
    assert compute(t['amount'][-1], x) == x['amount'][-1]


def test_count_string():
    s = symbol('name', 'var * ?string')
    x = np.array(['Alice', np.nan, 'Bob', 'Denis', 'Edith'], dtype='object')
    assert compute(s.count(), x) == 4


def test_reductions_on_recarray():
    assert compute(t.count(), x) == len(x)


def test_count_nan():
    t = symbol('t', '3 * ?real')
    x = np.array([1.0, np.nan, 2.0])
    assert compute(t.count(), x) == 2


def test_distinct():
    x = np.array([('Alice', 100),
                  ('Alice', -200),
                  ('Bob', 100),
                  ('Bob', 100)],
                dtype=[('name', 'S5'), ('amount', 'i8')])

    t = symbol('t', 'var * {name: string, amount: int64}')

    assert eq(compute(t['name'].distinct(), x),
              np.unique(x['name']))
    assert eq(compute(t.distinct(), x),
              np.unique(x))


def test_distinct_on_recarray():
    rec = pd.DataFrame(
        [[0, 1],
         [0, 2],
         [1, 1],
         [1, 2]],
        columns=('a', 'b'),
    ).to_records(index=False)

    s = symbol('s', discover(rec))
    assert (
        compute(s.distinct('a'), rec) ==
        pd.DataFrame(
            [[0, 1],
             [1, 1]],
            columns=('a', 'b'),
        ).to_records(index=False)
    ).all()


def test_distinct_on_structured_array():
    arr = np.array(
        [(0., 1.),
         (0., 2.),
         (1., 1.),
         (1., 2.)],
        dtype=[('a', 'f4'), ('b', 'f4')],
    )

    s = symbol('s', discover(arr))
    assert(
        compute(s.distinct('a'), arr) ==
        np.array([(0., 1.), (1., 1.)], dtype=arr.dtype)
    ).all()


def test_distinct_on_str():
    rec = pd.DataFrame(
        [['a', 'a'],
         ['a', 'b'],
         ['b', 'a'],
         ['b', 'b']],
        columns=('a', 'b'),
    ).to_records(index=False).astype([('a', '<U1'), ('b', '<U1')])

    s = symbol('s', discover(rec))
    assert (
        compute(s.distinct('a'), rec) ==
        pd.DataFrame(
            [['a', 'a'],
             ['b', 'a']],
            columns=('a', 'b'),
        ).to_records(index=False).astype([('a', '<U1'), ('b', '<U1')])
    ).all()


def test_sort():
    assert eq(compute(t.sort('amount'), x),
              np.sort(x, order='amount'))

    assert eq(compute(t.sort('amount', ascending=False), x),
              np.sort(x, order='amount')[::-1])

    assert eq(compute(t.sort(['amount', 'id']), x),
              np.sort(x, order=['amount', 'id']))

    assert eq(compute(t.amount.sort(), x),
              np.sort(x['amount']))


def test_head():
    assert eq(compute(t.head(2), x),
              x[:2])


def test_tail():
    assert eq(compute(t.tail(2), x),
              x[-2:])


def test_label():
    expected = x['amount'] * 10
    expected = np.array(expected, dtype=[('foo', 'i8')])
    assert eq(compute((t['amount'] * 10).label('foo'), x),
              expected)


def test_relabel():
    expected = np.array(x, dtype=[('ID', 'i8'), ('NAME', 'S7'), ('amount', 'i8')])
    result = compute(t.relabel({'name': 'NAME', 'id': 'ID'}), x)

    assert result.dtype.names == expected.dtype.names
    assert eq(result, expected)


def test_by():
    expr = by(t.amount > 0, count=t.id.count())
    result = compute(expr, x)
    assert set(map(tuple, into(list, result))) == set([(False, 2), (True, 3)])


def test_compute_up_field():
    assert eq(compute(t['name'], x), x['name'])


def test_compute_up_projection():
    assert eq(compute_up(t[['name', 'amount']], x), x[['name', 'amount']])


ax = np.arange(30, dtype='f4').reshape((5, 3, 2))

a = symbol('a', discover(ax))

def test_slice():
    inds = [0, slice(2), slice(1, 3), slice(None, None, 2), [1, 2, 3],
            (0, 1), (0, slice(1, 3)), (slice(0, 3), slice(3, 1, -1)),
            (0, [1, 2])]
    for s in inds:
        assert (compute(a[s], ax) == ax[s]).all()


def test_array_reductions():
    for axis in [None, 0, 1, (0, 1), (2, 1)]:
        assert eq(compute(a.sum(axis=axis), ax), ax.sum(axis=axis))
        assert eq(compute(a.std(axis=axis), ax), ax.std(axis=axis))


def test_array_reductions_with_keepdims():
    for axis in [None, 0, 1, (0, 1), (2, 1)]:
        assert eq(compute(a.sum(axis=axis, keepdims=True), ax),
                 ax.sum(axis=axis, keepdims=True))


def test_summary_on_ndarray():
    assert compute(summary(total=a.sum(), min=a.min()), ax) == \
            (ax.min(), ax.sum())

    result = compute(summary(total=a.sum(), min=a.min(), keepdims=True), ax)
    expected = np.array([(ax.min(), ax.sum())],
                        dtype=[('min', 'float32'), ('total', 'float64')])
    assert result.ndim == ax.ndim
    assert eq(expected, result)


def test_summary_on_ndarray_with_axis():
    for axis in [0, 1, (1, 0)]:
        expr = summary(total=a.sum(), min=a.min(), axis=axis)
        result = compute(expr, ax)

        shape, dtype = to_numpy(expr.dshape)
        expected = np.empty(shape=shape, dtype=dtype)
        expected['total'] = ax.sum(axis=axis)
        expected['min'] = ax.min(axis=axis)

        assert eq(result, expected)


def test_utcfromtimestamp():
    t = symbol('t', '1 * int64')
    data = np.array([0, 1])
    expected = np.array(['1970-01-01T00:00:00Z', '1970-01-01T00:00:01Z'],
                        dtype='M8[us]')
    assert eq(compute(t.utcfromtimestamp, data), expected)


def test_nelements_structured_array():
    assert compute(t.nelements(), x) == len(x)
    assert compute(t.nelements(keepdims=True), x) == (len(x),)


def test_nelements_array():
    t = symbol('t', '5 * 4 * 3 * float64')
    x = np.random.randn(*t.shape)
    result = compute(t.nelements(axis=(0, 1)), x)
    np.testing.assert_array_equal(result, np.array([20, 20, 20]))

    result = compute(t.nelements(axis=1), x)
    np.testing.assert_array_equal(result, 4 * np.ones((5, 3)))


def test_nrows():
    assert compute(t.nrows, x) == len(x)


dts = np.array(['2000-06-25T12:30:04Z', '2000-06-28T12:50:05Z'],
               dtype='M8[us]')
s = symbol('s', 'var * datetime')

def test_datetime_truncation():

    assert eq(compute(s.truncate(1, 'day'), dts),
              dts.astype('M8[D]'))
    assert eq(compute(s.truncate(2, 'seconds'), dts),
              np.array(['2000-06-25T12:30:04Z', '2000-06-28T12:50:04Z'],
                       dtype='M8[s]'))
    assert eq(compute(s.truncate(2, 'weeks'), dts),
              np.array(['2000-06-18', '2000-06-18'], dtype='M8[D]'))

    assert into(list, compute(s.truncate(1, 'week'), dts))[0].isoweekday() == 7


def test_hour():
    dts = [datetime(2000, 6, 20,  1, 00, 00),
           datetime(2000, 6, 20, 12, 59, 59),
           datetime(2000, 6, 20, 12, 00, 00),
           datetime(2000, 6, 20, 11, 59, 59)]
    dts = into(np.ndarray, dts)

    assert eq(compute(s.truncate(1, 'hour'), dts),
            into(np.ndarray, [datetime(2000, 6, 20,  1, 0),
                              datetime(2000, 6, 20, 12, 0),
                              datetime(2000, 6, 20, 12, 0),
                              datetime(2000, 6, 20, 11, 0)]))


def test_month():
    dts = [datetime(2000, 7, 1),
           datetime(2000, 6, 30),
           datetime(2000, 6, 1),
           datetime(2000, 5, 31)]
    dts = into(np.ndarray, dts)

    assert eq(compute(s.truncate(1, 'month'), dts),
            into(np.ndarray, [date(2000, 7, 1),
                              date(2000, 6, 1),
                              date(2000, 6, 1),
                              date(2000, 5, 1)]))


def test_truncate_on_np_datetime64_scalar():
    s = symbol('s', 'datetime')
    data = np.datetime64('2000-01-02T12:30:00Z')
    assert compute(s.truncate(1, 'day'), data) == data.astype('M8[D]')


def test_numpy_and_python_datetime_truncate_agree_on_start_of_week():
    s = symbol('s', 'datetime')
    n = np.datetime64('2014-11-11')
    p = datetime(2014, 11, 11)
    expr = s.truncate(1, 'week')
    assert compute(expr, n) == compute(expr, p)


def test_add_multiple_ndarrays():
    a = symbol('a', '5 * 4 * int64')
    b = symbol('b', '5 * 4 * float32')
    x = np.arange(9, dtype='int64').reshape(3, 3)
    y = (x + 1).astype('float32')
    expr = sin(a) + 2 * b
    scope = {a: x, b: y}
    expected = sin(x) + 2 * y

    # check that we cast correctly
    assert expr.dshape == dshape('5 * 4 * float64')

    np.testing.assert_array_equal(compute(expr, scope), expected)
    np.testing.assert_array_equal(compute(expr, scope, optimize=False),
                                  expected)


nA = np.arange(30, dtype='f4').reshape((5, 6))
ny = np.arange(6, dtype='f4')

A = symbol('A', discover(nA))
y = symbol('y', discover(ny))


def test_transpose():
    assert eq(compute(A.T, nA), nA.T)
    assert eq(compute(A.transpose((0, 1)), nA), nA)


def test_dot():
    assert eq(compute(y.dot(y), {y: ny}), np.dot(ny, ny))
    assert eq(compute(A.dot(y), {A: nA, y: ny}), np.dot(nA, ny))


def test_subexpr_datetime():
    data = pd.date_range(start='01/01/2010', end='01/04/2010', freq='D').values
    s = symbol('s', discover(data))
    result = compute(s.truncate(days=2).day, data)
    expected = np.array([31, 2, 2, 4])
    np.testing.assert_array_equal(result, expected)


def test_mixed_types():
    x = np.array([[(4, 180), (4, 184), (4, 188), (4, 192), (4, 196)],
                  [(4, 660), (4, 664), (4, 668), (4, 672), (4, 676)],
                  [(4, 1140), (4, 1144), (4, 1148), (4, 1152), (4, 1156)],
                  [(4, 1620), (4, 1624), (4, 1628), (4, 1632), (4, 1636)],
                  [(4, 2100), (4, 2104), (4, 2108), (4, 2112), (4, 2116)]],
                 dtype=[('count', '<i4'), ('total', '<i8')])
    aggregate = symbol('aggregate', discover(x))
    result = compute(aggregate.total.sum(axis=(0,)) /
                     aggregate['count'].sum(axis=(0,)), x)
    expected = (x['total'].sum(axis=0, keepdims=True) /
                x['count'].sum(axis=0, keepdims=True)).squeeze()
    np.testing.assert_array_equal(result, expected)


def test_broadcast_compute_against_numbers_and_arrays():
    A = symbol('A', '5 * float32')
    a = symbol('a', 'float32')
    b = symbol('b', 'float32')
    x = np.arange(5, dtype='f4')
    expr = Broadcast((A, b), (a, b), a + b)
    result = compute(expr, {A: x, b: 10})
    assert eq(result, x + 10)


def test_map():
    pytest.importorskip('numba')
    a = np.arange(10.0)
    f = lambda x: np.sin(x) + 1.03 * np.cos(x) ** 2
    x = symbol('x', discover(a))
    expr = x.map(f, 'float64')
    result = compute(expr, a)
    expected = f(a)

    # make sure we're not going to pandas here
    assert type(result) == np.ndarray
    assert type(result) == type(expected)

    np.testing.assert_array_equal(result, expected)


def test_vector_norm():
    x = np.arange(30).reshape((5, 6))
    s = symbol('x', discover(x))

    assert eq(compute(s.vnorm(), x),
              np.linalg.norm(x))
    assert eq(compute(s.vnorm(ord=1), x),
              np.linalg.norm(x.flatten(), ord=1))
    assert eq(compute(s.vnorm(ord=4, axis=0), x),
              np.linalg.norm(x, ord=4, axis=0))

    expr = s.vnorm(ord=4, axis=0, keepdims=True)
    assert expr.shape == compute(expr, x).shape


def test_join():
    cities = np.array([('Alice', 'NYC'),
                       ('Alice', 'LA'),
                       ('Bob', 'Chicago')],
                      dtype=[('name', 'S7'), ('city', 'O')])

    c = symbol('cities', discover(cities))

    expr = join(t, c, 'name')
    result = compute(expr, {t: x, c: cities})
    assert (b'Alice', 1, 100, 'LA') in into(list, result)


def test_query_with_strings():
    b = np.array([('a', 1), ('b', 2), ('c', 3)],
                 dtype=[('x', 'S1'), ('y', 'i4')])

    s = symbol('s', discover(b))
    assert compute(s[s.x == b'b'], b).tolist() == [(b'b', 2)]


@pytest.mark.parametrize('keys', [['a'], list('bc')])
def test_isin(keys):
    b = np.array([('a', 1), ('b', 2), ('c', 3), ('a', 4), ('c', 5), ('b', 6)],
                 dtype=[('x', 'S1'), ('y', 'i4')])

    s = symbol('s', discover(b))
    result = compute(s.x.isin(keys), b)
    expected = np.in1d(b['x'], keys)
    np.testing.assert_array_equal(result, expected)


def test_nunique_recarray():
    b = np.array([('a', 1), ('b', 2), ('c', 3), ('a', 4), ('c', 5), ('b', 6),
                  ('a', 1), ('b', 2)],
                 dtype=[('x', 'S1'), ('y', 'i4')])
    s = symbol('s', discover(b))
    expr = s.nunique()
    assert compute(expr, b) == len(np.unique(b))


def test_str_repeat():
    a = np.array(('a', 'b', 'c'))
    s = symbol('s', discover(a))
    expr = s.repeat(3)
    assert all(compute(expr, a) == np.char.multiply(a, 3))


def test_str_interp():
    a = np.array(('%s', '%s', '%s'))
    s = symbol('s', discover(a))
    expr = s.interp(1)
    assert all(compute(expr, a) == np.char.mod(a, 1))


def test_timedelta_arith():
    dates = np.arange('2014-01-01', '2014-02-01', dtype='datetime64')
    delta = np.timedelta64(1, 'D')
    sym = symbol('s', discover(dates))
    assert (compute(sym + delta, dates) == dates + delta).all()
    assert (compute(sym - delta, dates) == dates - delta).all()
    assert (
        compute(sym - (sym - delta), dates) ==
        dates - (dates - delta)
    ).all()


def test_coerce():
    x = np.arange(1, 3)
    s = symbol('s', discover(x))
    np.testing.assert_array_equal(compute(s.coerce('float64'), x),
                                  np.arange(1.0, 3.0))


def test_concat_arr():
    s_data = np.arange(15)
    t_data = np.arange(15, 30)

    s = symbol('s', discover(s_data))
    t = symbol('t', discover(t_data))

    assert (
        compute(concat(s, t), {s: s_data, t: t_data}) ==
        np.arange(30)
    ).all()


def test_concat_mat():
    s_data = np.arange(15).reshape(5, 3)
    t_data = np.arange(15, 30).reshape(5, 3)

    s = symbol('s', discover(s_data))
    t = symbol('t', discover(t_data))

    assert (
        compute(concat(s, t), {s: s_data, t: t_data}) ==
        np.arange(30).reshape(10, 3)
    ).all()
    assert (
        compute(concat(s, t, axis=1), {s: s_data, t: t_data}) ==
        np.concatenate((s_data, t_data), axis=1)
    ).all()


@pytest.mark.parametrize('dtype', ['int64', 'float64'])
def test_least(dtype):
    s_data = np.arange(15, dtype=dtype).reshape(5, 3)
    t_data = np.arange(15, 30, dtype=dtype).reshape(5, 3)
    s = symbol('s', discover(s_data))
    t = symbol('t', discover(t_data))
    expr = least(s, t)
    result = compute(expr, {s: s_data, t: t_data})
    expected = np.minimum(s_data, t_data)
    assert np.all(result == expected)


@pytest.mark.parametrize('dtype', ['int64', 'float64'])
def test_least_mixed(dtype):
    s_data = np.array([2, 1], dtype=dtype)
    t_data = np.array([1, 2], dtype=dtype)
    s = symbol('s', discover(s_data))
    t = symbol('t', discover(t_data))
    expr = least(s, t)
    result = compute(expr, {s: s_data, t: t_data})
    expected = np.minimum(s_data, t_data)
    assert np.all(result == expected)


@pytest.mark.parametrize('dtype', ['int64', 'float64'])
def test_greatest(dtype):
    s_data = np.arange(15, dtype=dtype).reshape(5, 3)
    t_data = np.arange(15, 30, dtype=dtype).reshape(5, 3)
    s = symbol('s', discover(s_data))
    t = symbol('t', discover(t_data))
    expr = greatest(s, t)
    result = compute(expr, {s: s_data, t: t_data})
    expected = np.maximum(s_data, t_data)
    assert np.all(result == expected)


@pytest.mark.parametrize('dtype', ['int64', 'float64'])
def test_greatest_mixed(dtype):
    s_data = np.array([2, 1], dtype=dtype)
    t_data = np.array([1, 2], dtype=dtype)
    s = symbol('s', discover(s_data))
    t = symbol('t', discover(t_data))
    expr = greatest(s, t)
    result = compute(expr, {s: s_data, t: t_data})
    expected = np.maximum(s_data, t_data)
    assert np.all(result == expected)


binary_name_map = {
    'atan2': 'arctan2'
}


@pytest.mark.parametrize(
    ['func', 'kwargs'],
    itertools.product(['copysign', 'ldexp'], [dict(optimize=False), dict()])
    )
def test_binary_math(func, kwargs):
    s_data = np.arange(15).reshape(5, 3)
    t_data = np.arange(15, 30).reshape(5, 3)
    s = symbol('s', discover(s_data))
    t = symbol('t', discover(t_data))
    scope = {s: s_data, t: t_data}
    result = compute(getattr(blaze, func)(s, t), scope, **kwargs)
    expected = getattr(np, binary_name_map.get(func, func))(s_data, t_data)
    np.testing.assert_equal(result, expected)


@pytest.mark.parametrize(
    ['func', 'kwargs'],
    itertools.product(['atan2', 'hypot'], [dict(optimize=False), dict()])
)
def test_floating_binary_math(func, kwargs):
    s_data = np.arange(15).reshape(5, 3)
    t_data = np.arange(15, 30).reshape(5, 3)
    s = symbol('s', discover(s_data))
    t = symbol('t', discover(t_data))
    scope = {s: s_data, t: t_data}
    result = compute(getattr(blaze, func)(s, t), scope, **kwargs)
    expected = getattr(np, binary_name_map.get(func, func))(s_data, t_data)
    np.testing.assert_allclose(result, expected)


def test_selection_inner_inputs():
    s_data = np.arange(5).reshape(5, 1)
    t_data = np.arange(5).reshape(5, 1)

    s = symbol('s', 'var * {a: int64}')
    t = symbol('t', 'var * {a: int64}')

    assert (
        compute(s[s.a == t.a], {s: s_data, t: t_data}) ==
        s_data
    ).all()


def test_coalesce():
    data = np.array([0, None, 1, None, 2, None])

    s = symbol('s', 'var * ?int')
    t = symbol('t', 'int')
    u = symbol('u', '?int')
    v = symbol('v', 'var * int')
    w = symbol('w', 'var * ?int')

    # array to scalar
    np.testing.assert_array_equal(
        compute(coalesce(s, t), {s: data, t: -1}),
        np.array([0, -1, 1, -1, 2, -1]),
    )
    # array to scalar with NULL
    np.testing.assert_array_equal(
        compute(coalesce(s, u), {s: data, u: None}),
        np.array([0, None, 1, None, 2, None], dtype=object)
    )
    # array to array
    np.testing.assert_array_equal(
        compute(coalesce(s, v), {
            s: data, v: np.array([-1, -2, -3, -4, -5, -6]),
        }),
        np.array([0, -2, 1, -4, 2, -6])
    )
    # array to array with NULL
    np.testing.assert_array_equal(
        compute(coalesce(s, w), {
            s: data, w: np.array([-1, None, -3, -4, -5, -6]),
        }),
        np.array([0, None, 1, -4, 2, -6]),
    )
